# cython: language_level=3
#
# Copyright 2015 Knowledge Economy Developments Ltd
# Copyright 2014 David Wells

# Henry Gomersall
# heng@kedevelopments.co.uk
#
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#
# * Redistributions of source code must retain the above copyright notice, this
# list of conditions and the following disclaimer.
#
# * Redistributions in binary form must reproduce the above copyright notice,
# this list of conditions and the following disclaimer in the documentation
# and/or other materials provided with the distribution.
#
# * Neither the name of the copyright holder nor the names of its contributors
# may be used to endorse or promote products derived from this software without
# specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
# ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE
# LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
# CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
# SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
# INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
# CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
# ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
# POSSIBILITY OF SUCH DAMAGE.
#
import numpy as np
cimport numpy as np
from libc.stdlib cimport calloc, malloc, free
from libc.stdint cimport intptr_t, int64_t
from libc cimport limits
import warnings
import threading

cdef int _simd_alignment = simd_alignment()

#: A tuple of simd alignments that make sense for this cpu
if _simd_alignment == 16:
    _valid_simd_alignments = (16,)

elif _simd_alignment == 32:
    _valid_simd_alignments = (16, 32)

else:
    _valid_simd_alignments = ()

cpdef byte_align(array, n=None, dtype=None):

    if not isinstance(array, np.ndarray):
        raise TypeError('Invalid array: byte_align requires a subclass '
                'of ndarray')

    if n is None:
        n = _simd_alignment

    if dtype is not None:
        if not array.dtype == dtype:
            update_dtype = True

    else:
        dtype = array.dtype
        update_dtype = False

    # See if we're already n byte aligned. If so, do nothing.
    offset = <intptr_t>np.PyArray_DATA(array) %n

    if offset is not 0 or update_dtype:

        _array_aligned = empty_aligned(array.shape, dtype, n=n)

        _array_aligned[:] = array

        array = _array_aligned.view(type=array.__class__)

    return array

cpdef is_byte_aligned(array, n=None):
    if not isinstance(array, np.ndarray):
        raise TypeError('Invalid array: is_n_byte_aligned requires a subclass '
                'of ndarray')

    if n is None:
        n = _simd_alignment

    # See if we're n byte aligned.
    offset = <intptr_t>np.PyArray_DATA(array) %n

    return not bool(offset)

cpdef empty_aligned(shape, dtype='float64', order='C', n=None):
    cdef long long array_length

    if n is None:
        n = _simd_alignment

    itemsize = np.dtype(dtype).itemsize

    # Apparently there is an issue with numpy.prod wrapping around on 32-bits
    # on Windows 64-bit. This shouldn't happen, but the following code
    # alleviates the problem.
    if not isinstance(shape, (int, np.integer)):
        array_length = 1
        for each_dimension in shape:
            array_length *= each_dimension

    else:
        array_length = shape

    # Allocate a new array that will contain the aligned data
    _array_aligned = np.empty(array_length*itemsize+n, dtype='int8')

    # We now need to know how to offset _array_aligned
    # so it is correctly aligned
    _array_aligned_offset = (n-<intptr_t>np.PyArray_DATA(_array_aligned))%n

    array = np.frombuffer(
            _array_aligned[_array_aligned_offset:_array_aligned_offset-n].data,
            dtype=dtype).reshape(shape, order=order)

    return array

cpdef zeros_aligned(shape, dtype='float64', order='C', n=None):
    array = empty_aligned(shape, dtype=dtype, order=order, n=n)
    array.fill(0)
    return array

cpdef ones_aligned(shape, dtype='float64', order='C', n=None):
    array = empty_aligned(shape, dtype=dtype, order=order, n=n)
    array.fill(1)
    return array

# the total number of types pyfftw can support
cdef int _n_types = 3
cdef object _all_types = ['32', '64']
_all_types_human_readable = {
    '32': 'single',
    '64': 'double'
}

cdef object directions
directions = {'FFTW_FORWARD': FFTW_FORWARD,
        'FFTW_BACKWARD': FFTW_BACKWARD,
        'FFTW_REDFT00': FFTW_REDFT00,
        'FFTW_REDFT10': FFTW_REDFT10,
        'FFTW_REDFT01': FFTW_REDFT01,
        'FFTW_REDFT11': FFTW_REDFT11,
        'FFTW_RODFT00': FFTW_RODFT00,
        'FFTW_RODFT10': FFTW_RODFT10,
        'FFTW_RODFT01': FFTW_RODFT01,
        'FFTW_RODFT11': FFTW_RODFT11}

cdef object directions_lookup
directions_lookup = {FFTW_FORWARD: 'FFTW_FORWARD',
        FFTW_BACKWARD: 'FFTW_BACKWARD',
        FFTW_REDFT00: 'FFTW_REDFT00',
        FFTW_REDFT10: 'FFTW_REDFT10',
        FFTW_REDFT01: 'FFTW_REDFT01',
        FFTW_REDFT11: 'FFTW_REDFT11',
        FFTW_RODFT00: 'FFTW_RODFT00',
        FFTW_RODFT10: 'FFTW_RODFT10',
        FFTW_RODFT01: 'FFTW_RODFT01',
        FFTW_RODFT11: 'FFTW_RODFT11'}

cdef object flag_dict
flag_dict = {'FFTW_MEASURE': FFTW_MEASURE,
        'FFTW_EXHAUSTIVE': FFTW_EXHAUSTIVE,
        'FFTW_PATIENT': FFTW_PATIENT,
        'FFTW_ESTIMATE': FFTW_ESTIMATE,
        'FFTW_UNALIGNED': FFTW_UNALIGNED,
        'FFTW_DESTROY_INPUT': FFTW_DESTROY_INPUT,
        'FFTW_WISDOM_ONLY': FFTW_WISDOM_ONLY}

# Need a global lock to protect FFTW planning so that multiple Python threads
# do not attempt to plan simultaneously.
cdef object plan_lock = threading.Lock()

# Function wrappers
# =================
# All of these have the same signature as the fftw_generic functions
# defined in the .pxd file. The arguments and return values are
# cast as required in order to call the actual fftw functions.
#
# The wrapper function names are simply the fftw names prefixed
# with a single underscore.

#     Planners
#     ========
#
# Complex double precision
cdef void* _fftw_plan_guru_dft(
            int rank, fftw_iodim *dims,
            int howmany_rank, fftw_iodim *howmany_dims,
            void *_in, void *_out,
            int *direction, unsigned flags) nogil:

    return <void *>fftw_plan_guru_dft(rank, dims,
            howmany_rank, howmany_dims,
            <cdouble *>_in, <cdouble *>_out,
            direction[0], flags)

# real to complex double precision
cdef void* _fftw_plan_guru_dft_r2c(
            int rank, fftw_iodim *dims,
            int howmany_rank, fftw_iodim *howmany_dims,
            void *_in, void *_out,
            int *direction, unsigned flags) nogil:

    return <void *>fftw_plan_guru_dft_r2c(rank, dims,
            howmany_rank, howmany_dims,
            <double *>_in, <cdouble *>_out,
            flags)

# complex to real double precision
cdef void* _fftw_plan_guru_dft_c2r(
            int rank, fftw_iodim *dims,
            int howmany_rank, fftw_iodim *howmany_dims,
            void *_in, void *_out,
            int *direction, unsigned flags) nogil:

    return <void *>fftw_plan_guru_dft_c2r(rank, dims,
            howmany_rank, howmany_dims,
            <cdouble *>_in, <double *>_out,
            flags)

# real to real double precision
cdef void* _fftw_plan_guru_r2r(
            int rank, fftw_iodim *dims,
            int howmany_rank, fftw_iodim *howmany_dims,
            void *_in, void *_out,
            int *direction, int flags):

    return <void *>fftw_plan_guru_r2r(rank, dims,
            howmany_rank, howmany_dims,
            <double *>_in, <double *>_out,
            direction, flags)

# Complex single precision
cdef void* _fftwf_plan_guru_dft(
            int rank, fftw_iodim *dims,
            int howmany_rank, fftw_iodim *howmany_dims,
            void *_in, void *_out,
            int *direction, unsigned flags) nogil:

    return <void *>fftwf_plan_guru_dft(rank, dims,
            howmany_rank, howmany_dims,
            <cfloat *>_in, <cfloat *>_out,
            direction[0], flags)

# real to complex single precision
cdef void* _fftwf_plan_guru_dft_r2c(
            int rank, fftw_iodim *dims,
            int howmany_rank, fftw_iodim *howmany_dims,
            void *_in, void *_out,
            int *direction, unsigned flags) nogil:

    return <void *>fftwf_plan_guru_dft_r2c(rank, dims,
            howmany_rank, howmany_dims,
            <float *>_in, <cfloat *>_out,
            flags)

# complex to real single precision
cdef void* _fftwf_plan_guru_dft_c2r(
            int rank, fftw_iodim *dims,
            int howmany_rank, fftw_iodim *howmany_dims,
            void *_in, void *_out,
            int *direction, unsigned flags) nogil:

    return <void *>fftwf_plan_guru_dft_c2r(rank, dims,
            howmany_rank, howmany_dims,
            <cfloat *>_in, <float *>_out,
            flags)

# real to real single precision
cdef void* _fftwf_plan_guru_r2r(
            int rank, fftw_iodim *dims,
            int howmany_rank, fftw_iodim *howmany_dims,
            void *_in, void *_out,
            int *direction, int flags):

    return <void *>fftwf_plan_guru_r2r(rank, dims,
            howmany_rank, howmany_dims,
            <float *>_in, <float *>_out,
            direction, flags)

#    Executors
#    =========
#
# Complex double precision
cdef void _fftw_execute_dft(void *_plan, void *_in, void *_out) nogil:

    fftw_execute_dft(<fftw_plan>_plan,
            <cdouble *>_in, <cdouble *>_out)

# real to complex double precision
cdef void _fftw_execute_dft_r2c(void *_plan, void *_in, void *_out) nogil:

    fftw_execute_dft_r2c(<fftw_plan>_plan,
            <double *>_in, <cdouble *>_out)

# complex to real double precision
cdef void _fftw_execute_dft_c2r(void *_plan, void *_in, void *_out) nogil:

    fftw_execute_dft_c2r(<fftw_plan>_plan,
            <cdouble *>_in, <double *>_out)

# Complex single precision
cdef void _fftwf_execute_dft(void *_plan, void *_in, void *_out) nogil:

    fftwf_execute_dft(<fftwf_plan>_plan,
            <cfloat *>_in, <cfloat *>_out)

# real to complex single precision
cdef void _fftwf_execute_dft_r2c(void *_plan, void *_in, void *_out) nogil:

    fftwf_execute_dft_r2c(<fftwf_plan>_plan,
            <float *>_in, <cfloat *>_out)

# complex to real single precision
cdef void _fftwf_execute_dft_c2r(void *_plan, void *_in, void *_out) nogil:

    fftwf_execute_dft_c2r(<fftwf_plan>_plan,
            <cfloat *>_in, <float *>_out)

# real to real double precision
cdef void _fftw_execute_r2r(void *_plan, void *_in, void *_out) nogil:

    fftw_execute_r2r(<fftw_plan>_plan, <double *>_in, <double *>_out)

# real to real single precision
cdef void _fftwf_execute_r2r(void *_plan, void *_in, void *_out) nogil:

    fftwf_execute_r2r(<fftwf_plan>_plan, <float *>_in, <float *>_out)

#    Destroyers
#    ==========
#
# Double precision
cdef void _fftw_destroy_plan(void *_plan):

    fftw_destroy_plan(<fftw_plan>_plan)

# Single precision
cdef void _fftwf_destroy_plan(void *_plan):

    fftwf_destroy_plan(<fftwf_plan>_plan)

# Function lookup tables
# ======================
# Planner table (of size the number of planners).
cdef fftw_generic_plan_guru planners[8]

cdef fftw_generic_plan_guru * _build_planner_list():
    planners[0] = <fftw_generic_plan_guru>&_fftw_plan_guru_dft
    planners[2] = <fftw_generic_plan_guru>&_fftw_plan_guru_dft_r2c
    planners[4] = <fftw_generic_plan_guru>&_fftw_plan_guru_dft_c2r
    planners[6] = <fftw_generic_plan_guru>&_fftw_plan_guru_r2r
    planners[1] = <fftw_generic_plan_guru>&_fftwf_plan_guru_dft
    planners[3] = <fftw_generic_plan_guru>&_fftwf_plan_guru_dft_r2c
    planners[5] = <fftw_generic_plan_guru>&_fftwf_plan_guru_dft_c2r
    planners[7] = <fftw_generic_plan_guru>&_fftwf_plan_guru_r2r

# Executor table (of size the number of executors)
cdef fftw_generic_execute executors[8]

cdef fftw_generic_execute * _build_executor_list():
    executors[0] = <fftw_generic_execute>&_fftw_execute_dft
    executors[2] = <fftw_generic_execute>&_fftw_execute_dft_r2c
    executors[4] = <fftw_generic_execute>&_fftw_execute_dft_c2r
    executors[6] = <fftw_generic_execute>&_fftw_execute_r2r
    executors[1] = <fftw_generic_execute>&_fftwf_execute_dft
    executors[3] = <fftw_generic_execute>&_fftwf_execute_dft_r2c
    executors[5] = <fftw_generic_execute>&_fftwf_execute_dft_c2r
    executors[7] = <fftw_generic_execute>&_fftwf_execute_r2r

# Destroyer table (of size the number of destroyers)
cdef fftw_generic_destroy_plan destroyers[2]

cdef fftw_generic_destroy_plan * _build_destroyer_list():
    destroyers[0] = <fftw_generic_destroy_plan>&_fftw_destroy_plan
    destroyers[1] = <fftw_generic_destroy_plan>&_fftwf_destroy_plan

# nthreads plan setters table
cdef fftw_generic_plan_with_nthreads nthreads_plan_setters[2]

cdef fftw_generic_plan_with_nthreads * _build_nthreads_plan_setters_list():
    nthreads_plan_setters[0] = (
        <fftw_generic_plan_with_nthreads>&fftw_plan_with_nthreads)
    nthreads_plan_setters[1] = (
        <fftw_generic_plan_with_nthreads>&fftwf_plan_with_nthreads)

# Set planner timelimits
cdef fftw_generic_set_timelimit set_timelimit_funcs[2]

cdef fftw_generic_set_timelimit * _build_set_timelimit_funcs_list():
    set_timelimit_funcs[0] = (
        <fftw_generic_set_timelimit>&fftw_set_timelimit)
    set_timelimit_funcs[1] = (
        <fftw_generic_set_timelimit>&fftwf_set_timelimit)

# Data validators table
cdef validator validators[2]

cdef validator * _build_validators_list():
    validators[0] = &_validate_r2c_arrays
    validators[1] = &_validate_c2r_arrays

# Validator functions
# ===================
cdef bint _validate_r2c_arrays(np.ndarray input_array,
        np.ndarray output_array, int64_t *axes, int64_t *not_axes,
        int64_t axes_length):
    ''' Validates the input and output array to check for
    a valid real to complex transform.
    '''
    # We firstly need to confirm that the dimenions of the arrays
    # are the same
    if not (input_array.ndim == output_array.ndim):
        return False

    in_shape = input_array.shape
    out_shape = output_array.shape

    for n in range(axes_length - 1):
        if not out_shape[axes[n]] == in_shape[axes[n]]:
            return False

    # The critical axis is the last of those over which the
    # FFT is taken.
    if not (out_shape[axes[axes_length-1]]
            == in_shape[axes[axes_length-1]]//2 + 1):
        return False

    for n in range(input_array.ndim - axes_length):
        if not out_shape[not_axes[n]] == in_shape[not_axes[n]]:
            return False

    return True

cdef bint _validate_c2r_arrays(np.ndarray input_array,
        np.ndarray output_array, int64_t *axes, int64_t *not_axes,
        int64_t axes_length):
    ''' Validates the input and output array to check for
    a valid complex to real transform.
    '''

    # We firstly need to confirm that the dimenions of the arrays
    # are the same
    if not (input_array.ndim == output_array.ndim):
        return False

    in_shape = input_array.shape
    out_shape = output_array.shape

    for n in range(axes_length - 1):
        if not in_shape[axes[n]] == out_shape[axes[n]]:
            return False

    # The critical axis is the last of those over which the
    # FFT is taken.
    if not (in_shape[axes[axes_length-1]]
            == out_shape[axes[axes_length-1]]//2 + 1):
        return False

    for n in range(input_array.ndim - axes_length):
        if not in_shape[not_axes[n]] == out_shape[not_axes[n]]:
            return False

    return True

# Shape lookup functions
# ======================
def _lookup_shape_r2c_arrays(input_array, output_array):
    return input_array.shape

def _lookup_shape_c2r_arrays(input_array, output_array):
    return output_array.shape

# fftw_schemes is a dictionary with a mapping from a keys,
# which are a tuple of the string representation of numpy
# dtypes to a scheme name.
#
# scheme_functions is a dictionary of functions, either
# an index to the array of functions in the case of
# 'planner', 'executor' and 'generic_precision' or a callable
# in the case of 'validator' (generic_precision is a catchall for
# functions that only change based on the precision changing -
# i.e the prefix fftw, fftwl and fftwf is the only bit that changes).
#
# The array indices refer to the relevant functions for each scheme,
# the tables to which are defined above.
#
# The 'validator' function is a callable for validating the arrays
# that has the following signature:
# bool callable(ndarray in_array, ndarray out_array, axes, not_axes)
# and checks that the arrays are a valid pair. If it is set to None,
# then the default check is applied, which confirms that the arrays
# have the same shape.
#
# The 'fft_shape_lookup' function is a callable for returning the
# FFT shape - that is, an array that describes the length of the
# fft along each axis. It has the following signature:
# fft_shape = fft_shape_lookup(in_array, out_array)
# (note that this does not correspond to the lengths of the FFT that is
# actually taken, it's the lengths of the FFT that *could* be taken
# along each axis. It's necessary because the real FFT has a length
# that is different to the length of the input array).

cdef object fftw_schemes
fftw_schemes = {
        (np.dtype('complex128'), np.dtype('complex128')): ('c2c', '64'),
        (np.dtype('complex64'), np.dtype('complex64')): ('c2c', '32'),
        (np.dtype('float64'), np.dtype('complex128')): ('r2c', '64'),
        (np.dtype('float32'), np.dtype('complex64')): ('r2c', '32'),
        (np.dtype('complex128'), np.dtype('float64')): ('c2r', '64'),
        (np.dtype('complex64'), np.dtype('float32')): ('c2r', '32'),
        (np.dtype('float32'), np.dtype('float32')): ('r2r', '32'),
        (np.dtype('float64'), np.dtype('float64')): ('r2r', '64')}

cdef object fftw_default_output
fftw_default_output = {
    np.dtype('float32'): np.dtype('complex64'),
    np.dtype('float64'): np.dtype('complex128'),
    np.dtype('complex64'): np.dtype('complex64'),
    np.dtype('complex128'): np.dtype('complex128')}

cdef object scheme_directions
scheme_directions = {
        ('c2c', '64'): ['FFTW_FORWARD', 'FFTW_BACKWARD'],
        ('c2c', '32'): ['FFTW_FORWARD', 'FFTW_BACKWARD'],
        ('r2c', '64'): ['FFTW_FORWARD'],
        ('r2c', '32'): ['FFTW_FORWARD'],
        ('c2r', '64'): ['FFTW_BACKWARD'],
        ('c2r', '32'): ['FFTW_BACKWARD'],
        ('r2r', '64'): ['FFTW_REDFT00', 'FFTW_REDFT10', 'FFTW_REDFT01',
                        'FFTW_REDFT11', 'FFTW_RODFT00', 'FFTW_RODFT10',
                        'FFTW_RODFT01', 'FFTW_RODFT11'],
        ('r2r', '32'): ['FFTW_REDFT00', 'FFTW_REDFT10', 'FFTW_REDFT01',
                        'FFTW_REDFT11', 'FFTW_RODFT00', 'FFTW_RODFT10',
                        'FFTW_RODFT01', 'FFTW_RODFT11']}

# In the following, -1 denotes using the default. A segfault has been
# reported on some systems when this is set to None. It seems
# sufficiently trivial to use -1 in place of None, especially given
# that scheme_functions is an internal cdef object.
cdef object _scheme_functions = {}

_scheme_functions.update({
('c2c', '64'): {'planner': 0, 'executor': 0, 'generic_precision': 0,
    'validator': -1, 'fft_shape_lookup': -1},
('r2c', '64'): {'planner': 2, 'executor': 2, 'generic_precision': 0,
    'validator': 0,
    'fft_shape_lookup': _lookup_shape_r2c_arrays},
('c2r', '64'): {'planner': 4, 'executor': 4, 'generic_precision': 0,
    'validator': 1,
    'fft_shape_lookup': _lookup_shape_c2r_arrays},
('r2r', '64'): {'planner': 6, 'executor': 6, 'generic_precision': 0,
    'validator': -1, 'fft_shape_lookup': -1}})

_scheme_functions.update({
('c2c', '32'): {'planner': 1, 'executor': 1, 'generic_precision': 1,
    'validator': -1, 'fft_shape_lookup': -1},
('r2c', '32'): {'planner': 3, 'executor': 3, 'generic_precision': 1,
    'validator': 0,
    'fft_shape_lookup': _lookup_shape_r2c_arrays},
('c2r', '32'): {'planner': 5, 'executor': 5, 'generic_precision': 1,
    'validator': 1,
    'fft_shape_lookup': _lookup_shape_c2r_arrays},
('r2r', '32'): {'planner': 7, 'executor': 7, 'generic_precision': 1,
    'validator': -1, 'fft_shape_lookup': -1}})

def scheme_functions(scheme):
    if scheme in _scheme_functions:
        return _scheme_functions[scheme]
    else:
        msg = "The scheme '%s' is not supported." % str(scheme)
        if scheme[1] in _all_types:
            msg += "\nRebuild pyFFTW with support for %s precision!" % \
                   _all_types_human_readable[scheme[1]]
        raise NotImplementedError(msg)

# Set the cleanup routine
cdef void _cleanup():
    fftw_cleanup()
    fftwf_cleanup()
    fftw_cleanup_threads()
    fftwf_cleanup_threads()

# Initialize the module

# Define the functions
_build_planner_list()
_build_destroyer_list()
_build_executor_list()
_build_nthreads_plan_setters_list()
_build_validators_list()
_build_set_timelimit_funcs_list()

fftw_init_threads()
fftwf_init_threads()

Py_AtExit(_cleanup)

# Helper functions
cdef void make_axes_unique(int64_t *axes, int64_t axes_length,
        int64_t **unique_axes, int64_t **not_axes, int64_t dimensions,
        int64_t *unique_axes_length):
    ''' Takes an array of axes and makes that array unique, returning
    the unique array in unique_axes. It also creates and fills another
    array, not_axes, with those axes that are not included in unique_axes.

    unique_axes_length is updated with the length of unique_axes.

    dimensions is the number of dimensions to which the axes array
    might refer.

    It is the responsibility of the caller to free unique_axes and not_axes.
    '''

    cdef int64_t unique_axes_count = 0
    cdef int64_t holding_offset = 0

    cdef int64_t *axes_holding = (
            <int64_t *>calloc(dimensions, sizeof(int64_t)))
    cdef int64_t *axes_holding_offset = (
            <int64_t *>calloc(dimensions, sizeof(int64_t)))

    for n in range(dimensions):
        axes_holding[n] = -1

    # Iterate over all the axes and store each index if it hasn't already
    # been stored (this keeps one and only one and the first index to axes
    # i.e. storing the unique set of entries).
    #
    # axes_holding_offset holds the shift due to repeated axes
    for n in range(axes_length):
        if axes_holding[axes[n]] == -1:
            axes_holding[axes[n]] = n
            axes_holding_offset[axes[n]] = holding_offset
            unique_axes_count += 1
        else:
            holding_offset += 1

    unique_axes[0] = <int64_t *>malloc(
            unique_axes_count * sizeof(int64_t))

    not_axes[0] = <int64_t *>malloc(
            (dimensions - unique_axes_count) * sizeof(int64_t))

    # Now we need to write back the unique axes to a tmp axes
    cdef int64_t not_axes_count = 0

    for n in range(dimensions):
        if axes_holding[n] != -1:
            unique_axes[0][axes_holding[n] - axes_holding_offset[n]] = (
                    axes[axes_holding[n]])

        else:
            not_axes[0][not_axes_count] = n
            not_axes_count += 1

    free(axes_holding)
    free(axes_holding_offset)

    unique_axes_length[0] = unique_axes_count


# The External Interface
# ======================
#
cdef class FFTW:
    def _get_N(self):
        return self._total_size

    N = property(_get_N)

    def _get_simd_aligned(self):
        return self._simd_allowed

    simd_aligned = property(_get_simd_aligned)

    def _get_input_alignment(self):
        return self._input_array_alignment

    input_alignment = property(_get_input_alignment)

    def _get_output_alignment(self):
        return self._output_array_alignment

    output_alignment = property(_get_output_alignment)

    def _get_flags_used(self):
        return tuple(self._flags_used)

    flags = property(_get_flags_used)

    def _get_input_array(self):
        return self._input_array

    input_array = property(_get_input_array)

    def _get_output_array(self):
        return self._output_array

    output_array = property(_get_output_array)

    def _get_input_strides(self):
        return self._input_strides

    input_strides = property(_get_input_strides)

    def _get_output_strides(self):
        return self._output_strides

    output_strides = property(_get_output_strides)

    def _get_input_shape(self):
        return self._input_shape

    input_shape = property(_get_input_shape)

    def _get_output_shape(self):
        return self._output_shape

    output_shape = property(_get_output_shape)

    def _get_input_dtype(self):
        return self._input_dtype

    input_dtype = property(_get_input_dtype)

    def _get_output_dtype(self):
        return self._output_dtype

    output_dtype = property(_get_output_dtype)

    def _get_direction(self):
        cdef int i
        transform_directions = list()
        if self._direction[0] in [FFTW_FORWARD, FFTW_BACKWARD]:
            # It would be nice to return a length-one list here (so that the
            # return type is always [str]). This is an annoying type difference,
            # but is backwards compatible.
            return directions_lookup[self._direction[0]]
        else:
            for i in range(self._rank):
                transform_directions.append(directions_lookup[
                        self._direction[i]])
        return transform_directions

    direction = property(_get_direction)

    def _get_axes(self):
        axes = []
        for i in range(self._rank):
            axes.append(self._axes[i])

        return tuple(axes)

    axes = property(_get_axes)

    def _get_normalise_idft(self):
        return self._normalise_idft

    normalise_idft = property(_get_normalise_idft)

    def _get_ortho(self):
        return self._ortho

    ortho = property(_get_ortho)

    def __cinit__(self, input_array, output_array, axes=(-1,),
                  direction='FFTW_FORWARD', flags=('FFTW_MEASURE',),
                  unsigned int threads=1, planning_timelimit=None,
                  bint normalise_idft=True, bint ortho=False, *args,
                  **kwargs):

        if isinstance(direction, str):
            given_directions = [direction]
        else:
            given_directions = list(direction)

        # Initialise the pointers that need to be freed
        self._plan = NULL
        self._dims = NULL
        self._howmany_dims = NULL

        self._axes = NULL
        self._not_axes = NULL
        self._direction = NULL

        self._normalise_idft = normalise_idft
        self._ortho = ortho
        if self._ortho and self._normalise_idft:
            raise ValueError('Invalid options: '
                'ortho and normalise_idft cannot both be True.')

        flags = list(flags)

        cdef double _planning_timelimit
        if planning_timelimit is None:
            _planning_timelimit = FFTW_NO_TIMELIMIT
        else:
            try:
                _planning_timelimit = planning_timelimit
            except TypeError:
                raise TypeError('Invalid planning timelimit: '
                        'The planning timelimit needs to be a float.')

        if not isinstance(input_array, np.ndarray):
            raise ValueError('Invalid input array: '
                    'The input array needs to be an instance '
                    'of numpy.ndarray')

        if not isinstance(output_array, np.ndarray):
            raise ValueError('Invalid output array: '
                    'The output array needs to be an instance '
                    'of numpy.ndarray')

        try:
            input_dtype = input_array.dtype
            output_dtype = output_array.dtype
            scheme = fftw_schemes[(input_dtype, output_dtype)]
        except KeyError:
            raise ValueError('Invalid scheme: '
                    'The output array and input array dtypes '
                    'do not correspond to a valid fftw scheme.')

        self._input_dtype = input_dtype
        self._output_dtype = output_dtype

        functions = scheme_functions(scheme)

        self._fftw_planner = planners[functions['planner']]
        self._fftw_execute = executors[functions['executor']]
        self._fftw_destroy = destroyers[functions['generic_precision']]

        self._nthreads_plan_setter = (
                nthreads_plan_setters[functions['generic_precision']])

        cdef fftw_generic_set_timelimit set_timelimit_func = (
                set_timelimit_funcs[functions['generic_precision']])

        # We're interested in the natural alignment on the real type, not
        # necessarily on the complex type At least one bug was found where
        # numpy reported an alignment on a complex dtype that was different
        # to that on the real type.
        cdef int natural_input_alignment = input_array.real.dtype.alignment
        cdef int natural_output_alignment = output_array.real.dtype.alignment

        # If either of the arrays is not aligned on a 16-byte boundary,
        # we set the FFTW_UNALIGNED flag. This disables SIMD.
        # (16 bytes is assumed to be the minimal alignment)
        if 'FFTW_UNALIGNED' in flags:
            self._simd_allowed = False
            self._input_array_alignment = natural_input_alignment
            self._output_array_alignment = natural_output_alignment

        else:

            self._input_array_alignment = -1
            self._output_array_alignment = -1

            for each_alignment in _valid_simd_alignments:
                if (<intptr_t>np.PyArray_DATA(input_array) %
                        each_alignment == 0 and
                        <intptr_t>np.PyArray_DATA(output_array) %
                        each_alignment == 0):

                    self._simd_allowed = True

                    self._input_array_alignment = each_alignment
                    self._output_array_alignment = each_alignment

                    break

            if (self._input_array_alignment == -1 or
                    self._output_array_alignment == -1):

                self._simd_allowed = False

                self._input_array_alignment = (
                        natural_input_alignment)
                self._output_array_alignment = (
                        natural_output_alignment)
                flags.append('FFTW_UNALIGNED')

        if (not (<intptr_t>np.PyArray_DATA(input_array)
            % self._input_array_alignment == 0)):
            raise ValueError('Invalid input alignment: '
                    'The input array is expected to lie on a %d '
                    'byte boundary.' % self._input_array_alignment)

        if (not (<intptr_t>np.PyArray_DATA(output_array)
            % self._output_array_alignment == 0)):
            raise ValueError('Invalid output alignment: '
                    'The output array is expected to lie on a %d '
                    'byte boundary.' % self._output_array_alignment)

        for direction in given_directions:
            if direction not in scheme_directions[scheme]:
                raise ValueError('Invalid direction: '
                        'The direction is not valid for the scheme. '
                        'Try setting it explicitly if it is not already.')

        self._direction = <int *>malloc(len(axes)*sizeof(int))

        real_transforms = True
        cdef int i
        if given_directions[0] in ['FFTW_FORWARD', 'FFTW_BACKWARD']:
            self._direction[0] = directions[given_directions[0]]
            real_transforms = False
        else:
            if len(axes) != len(given_directions):
                raise ValueError('For real-to-real transforms, there must '
                        'be exactly one specified transform for each '
                        'transformed axis.')
            for i in range(len(axes)):
                if given_directions[0] in ['FFTW_FORWARD', 'FFTW_BACKWARD']:
                    raise ValueError('Heterogeneous transforms cannot be '
                            'assigned with \'FFTW_FORWARD\' or '
                            '\'FFTW_BACKWARD\'.')
                else:
                    self._direction[i] = directions[given_directions[i]]

        self._input_shape = input_array.shape
        self._output_shape = output_array.shape

        self._input_array = input_array
        self._output_array = output_array

        self._axes = <int64_t *>malloc(len(axes)*sizeof(int64_t))
        for n in range(len(axes)):
            self._axes[n] = axes[n]

        # Set the negative entries to their actual index (use the size
        # of the shape array for this)
        cdef int64_t array_dimension = len(self._input_shape)

        for n in range(len(axes)):
            if self._axes[n] < 0:
                self._axes[n] = self._axes[n] + array_dimension

            if self._axes[n] >= array_dimension or self._axes[n] < 0:
                raise IndexError('Invalid axes: '
                    'The axes list cannot contain invalid axes.')

        cdef int64_t unique_axes_length
        cdef int64_t *unique_axes
        cdef int64_t *not_axes

        make_axes_unique(self._axes, len(axes), &unique_axes,
                &not_axes, array_dimension, &unique_axes_length)

        # and assign axes and not_axes to the filled arrays
        free(self._axes)
        self._axes = unique_axes
        self._not_axes = not_axes

        total_N = 1
        for n in range(unique_axes_length):
            if self._input_shape[self._axes[n]] == 0:
                raise ValueError('Zero length array: '
                    'The input array should have no zero length'
                    'axes over which the FFT is to be taken')

            if real_transforms:
                if self._direction[n] == FFTW_RODFT00:
                    total_N *= 2*(self._input_shape[self._axes[n]] + 1)
                elif self._direction[n] == FFTW_REDFT00:
                    if (self._input_shape[self._axes[n]] < 2):
                        raise ValueError('FFTW_REDFT00 (also known as DCT-1) is'
                                ' not defined for inputs of length less than two.')
                    total_N *= 2*(self._input_shape[self._axes[n]] - 1)
                else:
                    total_N *= 2*self._input_shape[self._axes[n]]
            else:
                if self._direction[0] == FFTW_FORWARD:
                    total_N *= self._input_shape[self._axes[n]]
                else:
                    total_N *= self._output_shape[self._axes[n]]

        self._total_size = total_N
        self._normalisation_scaling = 1/float(self.N)
        self._sqrt_normalisation_scaling = np.sqrt(self._normalisation_scaling)

        # Now we can validate the array shapes
        cdef validator _validator

        if functions['validator'] == -1:
            if not (output_array.shape == input_array.shape):
                raise ValueError('Invalid shapes: '
                        'The output array should be the same shape as the '
                        'input array for the given array dtypes.')
        else:
            _validator = validators[functions['validator']]
            if not _validator(input_array, output_array,
                    self._axes, self._not_axes, unique_axes_length):
                raise ValueError('Invalid shapes: '
                        'The input array and output array are invalid '
                        'complementary shapes for their dtypes.')

        self._rank = unique_axes_length
        self._howmany_rank = self._input_array.ndim - unique_axes_length

        self._flags = 0
        self._flags_used = []
        for each_flag in flags:
            try:
                self._flags |= flag_dict[each_flag]
                self._flags_used.append(each_flag)
            except KeyError:
                raise ValueError('Invalid flag: ' + '\'' +
                        each_flag + '\' is not a valid planner flag.')


        if ('FFTW_DESTROY_INPUT' not in flags) and (
                (scheme[0] != 'c2r') or not self._rank > 1):
            # The default in all possible cases is to preserve the input
            # This is not possible for r2c arrays with rank > 1
            self._flags |= FFTW_PRESERVE_INPUT

        # Set up the arrays of structs for holding the stride shape
        # information
        self._dims = <_fftw_iodim *>malloc(
                self._rank * sizeof(_fftw_iodim))
        self._howmany_dims = <_fftw_iodim *>malloc(
                self._howmany_rank * sizeof(_fftw_iodim))

        if self._dims == NULL or self._howmany_dims == NULL:
            # Not much else to do than raise an exception
            raise MemoryError

        # Find the strides for all the axes of both arrays in terms of the
        # number of items (as opposed to the number of bytes).
        self._input_strides = input_array.strides
        self._input_item_strides = tuple([stride/input_array.itemsize
            for stride in input_array.strides])
        self._output_strides = output_array.strides
        self._output_item_strides = tuple([stride/output_array.itemsize
            for stride in output_array.strides])

        # Make sure that the arrays are not too big for fftw
        # This is hard to test, so we cross our fingers and hope for the
        # best (any suggestions, please get in touch).
        for i in range(0, len(self._input_shape)):
            if self._input_shape[i] >= <Py_ssize_t> limits.INT_MAX:
                raise ValueError('Dimensions of the input array must be ' +
                        'less than ', str(limits.INT_MAX))

            if self._input_item_strides[i] >= <Py_ssize_t> limits.INT_MAX:
                raise ValueError('Strides of the input array must be ' +
                        'less than ', str(limits.INT_MAX))

        for i in range(0, len(self._output_shape)):
            if self._output_shape[i] >= <Py_ssize_t> limits.INT_MAX:
                raise ValueError('Dimensions of the output array must be ' +
                        'less than ', str(limits.INT_MAX))

            if self._output_item_strides[i] >= <Py_ssize_t> limits.INT_MAX:
                raise ValueError('Strides of the output array must be ' +
                        'less than ', str(limits.INT_MAX))

        fft_shape_lookup = functions['fft_shape_lookup']
        if fft_shape_lookup == -1:
            fft_shape = self._input_shape
        else:
            fft_shape = fft_shape_lookup(input_array, output_array)

        # Fill in the stride and shape information
        input_strides_array = self._input_item_strides
        output_strides_array = self._output_item_strides
        for i in range(0, self._rank):
            self._dims[i]._n = fft_shape[self._axes[i]]
            self._dims[i]._is = input_strides_array[self._axes[i]]
            self._dims[i]._os = output_strides_array[self._axes[i]]

        for i in range(0, self._howmany_rank):
            self._howmany_dims[i]._n = fft_shape[self._not_axes[i]]
            self._howmany_dims[i]._is = input_strides_array[self._not_axes[i]]
            self._howmany_dims[i]._os = output_strides_array[self._not_axes[i]]

        # parallel execution
        self._use_threads = (threads > 1)

        ## Point at which FFTW calls are made
        ## (and none should be made before this)

        # noop if threads library not available
        self._nthreads_plan_setter(threads)

        # Set the timelimit
        set_timelimit_func(_planning_timelimit)

        # Finally, construct the plan, after acquiring the global planner lock
        # (so that only one python thread can plan at a time, as the FFTW
        # planning functions are not thread-safe)

        # no self-lookups allowed in nogil block, so must grab all these first
        cdef void *plan
        cdef fftw_generic_plan_guru fftw_planner = self._fftw_planner
        cdef int rank = self._rank
        cdef fftw_iodim *dims = <fftw_iodim *>self._dims
        cdef int howmany_rank = self._howmany_rank
        cdef fftw_iodim *howmany_dims = <fftw_iodim *>self._howmany_dims
        cdef void *_in = <void *>np.PyArray_DATA(self._input_array)
        cdef void *_out = <void *>np.PyArray_DATA(self._output_array)
        cdef unsigned c_flags = self._flags

        with plan_lock, nogil:
            plan = fftw_planner(rank, dims, howmany_rank, howmany_dims,
                                _in, _out, self._direction, c_flags)
        self._plan = plan

        if self._plan == NULL:
            if 'FFTW_WISDOM_ONLY' in flags:
                raise RuntimeError('No FFTW wisdom is known for this plan.')
            else:
                raise RuntimeError('The data has an uncaught error that led '+
                    'to the planner returning NULL. This is a bug.')

    def __dealloc__(self):

        if not self._axes == NULL:
            free(self._axes)

        if not self._not_axes == NULL:
            free(self._not_axes)

        if not self._plan == NULL:
            self._fftw_destroy(self._plan)

        if not self._dims == NULL:
            free(self._dims)

        if not self._howmany_dims == NULL:
            free(self._howmany_dims)

        if not self._direction == NULL:
            free(self._direction)

    def __call__(self, input_array=None, output_array=None,
                 normalise_idft=None, ortho=None):
        if ortho is None:
            ortho = self._ortho
        if normalise_idft is None:
            normalise_idft = self._normalise_idft

        if ortho and normalise_idft:
            raise ValueError('Invalid options: ortho and normalise_idft cannot'
                             ' both be True.')

        if input_array is not None or output_array is not None:

            if input_array is None:
                input_array = self._input_array

            if output_array is None:
                output_array = self._output_array

            if not isinstance(input_array, np.ndarray):
                copy_needed = True
            elif (not input_array.dtype == self._input_dtype):
                copy_needed = True
            elif (not input_array.strides == self._input_strides):
                copy_needed = True
            elif not (<intptr_t>np.PyArray_DATA(input_array)
                    % self.input_alignment == 0):
                copy_needed = True
            else:
                copy_needed = False

            if copy_needed:

                if not isinstance(input_array, np.ndarray):
                    input_array = np.asanyarray(input_array)

                if not input_array.shape == self._input_shape:
                    raise ValueError('Invalid input shape: '
                            'The new input array should be the same shape '
                            'as the input array used to instantiate the '
                            'object.')

                self._input_array[:] = input_array

                if output_array is not None:
                    # No point wasting time if no update is necessary
                    # (which the copy above may have avoided)
                    input_array = self._input_array
                    self.update_arrays(input_array, output_array)

            else:
                self.update_arrays(input_array, output_array)

        self.execute()

        if ortho == True:
            self._output_array *= self._sqrt_normalisation_scaling

        if self._direction[0] == FFTW_BACKWARD and normalise_idft:
            self._output_array *= self._normalisation_scaling

        return self._output_array

    def update_arrays(self, new_input_array, new_output_array):
        if not isinstance(new_input_array, np.ndarray):
            raise ValueError('Invalid input array: '
                    'The new input array needs to be an instance '
                    'of numpy.ndarray')

        if not isinstance(new_output_array, np.ndarray):
            raise ValueError('Invalid output array '
                    'The new output array needs to be an instance '
                    'of numpy.ndarray')

        if not (<intptr_t>np.PyArray_DATA(new_input_array) %
                self.input_alignment == 0):
            raise ValueError('Invalid input alignment: '
                    'The original arrays were %d-byte aligned. It is '
                    'necessary that the update input array is similarly '
                    'aligned.' % self.input_alignment)

        if not (<intptr_t>np.PyArray_DATA(new_output_array) %
                self.output_alignment == 0):
            raise ValueError('Invalid output alignment: '
                    'The original arrays were %d-byte aligned. It is '
                    'necessary that the update output array is similarly '
                    'aligned.' % self.output_alignment)

        if not new_input_array.dtype == self._input_dtype:
            raise ValueError('Invalid input dtype: '
                    'The new input array is not of the same '
                    'dtype as was originally planned for.')

        if not new_output_array.dtype == self._output_dtype:
            raise ValueError('Invalid output dtype: '
                    'The new output array is not of the same '
                    'dtype as was originally planned for.')

        new_input_shape = new_input_array.shape
        new_output_shape = new_output_array.shape

        new_input_strides = new_input_array.strides
        new_output_strides = new_output_array.strides

        if not new_input_shape == self._input_shape:
            raise ValueError('Invalid input shape: '
                    'The new input array should be the same shape as '
                    'the input array used to instantiate the object.')

        if not new_output_shape == self._output_shape:
            raise ValueError('Invalid output shape: '
                    'The new output array should be the same shape as '
                    'the output array used to instantiate the object.')

        if not new_input_strides == self._input_strides:
            raise ValueError('Invalid input striding: '
                    'The strides should be identical for the new '
                    'input array as for the old.')

        if not new_output_strides == self._output_strides:
            raise ValueError('Invalid output striding: '
                    'The strides should be identical for the new '
                    'output array as for the old.')

        self._update_arrays(new_input_array, new_output_array)

    def execute(self):
        self._execute()



import os
import shutil
from datetime import datetime
from typing import List, Tuple
import pytest
import numpy as np
import pyrost as rst
from pyrost import simulation as st_sim

@pytest.fixture(params=[{'detx_size': 300, 'dety_size': 300, 'n_frames': 50, 'p0': 1e6,
                         'pix_size': 300, 'bar_size': 0.3, 'bar_rnd': 0.5, 'alpha': 0.05,
                         'num_threads': 4},
                        {'detx_size': 300, 'dety_size': 300, 'n_frames': 50, 'p0': 1e6,
                         'pix_size': 300, 'bar_size': 0.3, 'bar_rnd': 0.5, 'alpha': 0.03,
                         'num_threads': 4}],
                scope='session')
def st_params(request) -> st_sim.STParams:
    """Return a default instance of simulation parameters.
    """
    return st_sim.STParams.import_default(**request.param)

@pytest.fixture(scope='session')
def st_converter(st_params: st_sim.STParams) -> st_sim.STConverter:
    sim_obj = st_sim.STSim(st_params)
    ptych = sim_obj.ptychograph()
    return st_sim.STConverter(sim_obj, ptych)

@pytest.fixture(scope='session')
def temp_dir() -> str:
    now = datetime.now()
    path = now.strftime("temp_%m_%d_%H%M%S")
    os.mkdir(path)
    yield path
    shutil.rmtree(path)

@pytest.fixture(scope='function')
def ini_path(temp_dir: str) -> str:
    """Return a path to the experimental speckle tracking data.
    """
    path = os.path.join(temp_dir, 'test.ini')
    yield path
    os.remove(path)

@pytest.fixture
def crop(roi: Tuple[int, int, int, int]) -> rst.Crop:
    return rst.Crop(roi)

@pytest.fixture
def kamzik_converter(scan_num: int) -> rst.KamzikConverter:
    converter = rst.KamzikConverter()
    converter = converter.read_logs(f'/gpfs/cfel/group/cxi/labs/MLL-Sigray/scan-logs/Scan_{scan_num:d}.log')
    return converter

@pytest.fixture
def input_file(scan_num: int) -> rst.CXIStore:
    data_dir = f'/gpfs/cfel/group/cxi/labs/MLL-Sigray/scan-frames/Scan_{scan_num:d}'
    data_files = sorted([os.path.join(data_dir, path) for path in os.listdir(data_dir)
                         if path.endswith('Lambda.nxs')])
    return rst.CXIStore(data_files)

@pytest.fixture
def good_frames_list(good_frames: Tuple[int, int]) -> np.ndarray:
    return np.arange(good_frames[0], good_frames[1])

@pytest.mark.st_sim
def test_st_params(st_params: st_sim.STParams, ini_path: str):
    assert not os.path.isfile(ini_path)
    ini_parser = st_params.export_ini()
    with open(ini_path, 'w') as ini_file:
        ini_parser.write(ini_file)
    new_params = st_sim.STParams.import_ini(ini_path)
    assert new_params.export_dict() == st_params.export_dict()

@pytest.mark.st_sim
def test_save_and_load_sim(st_converter: st_sim.STConverter, temp_dir: str):
    assert os.path.isdir(temp_dir)
    out_path = os.path.join(temp_dir, 'sim.cxi')
    st_converter.save(out_path)
    assert os.path.isfile(out_path)
    data = st_converter.export_data(out_path).load()
    for attr in data.input_file:
        assert attr in data.contents()

@pytest.mark.standalone
def test_st_update_sim(st_converter: st_sim.STConverter, temp_dir: str):
    out_path = os.path.join(temp_dir, 'sim.cxi')
    data = st_converter.export_data(out_path).load()
    st_obj = data.get_st()
    h0 = st_obj.find_hopt(verbose=True)
    st_res = st_obj.train(search_window=(0.0, 10.0, 0.1), h0=h0, blur=8.0,
                          f_tol=-1.0, n_iter=10, verbose=True)
    data.import_st(st_res)
    fit_obj = data.get_fit(axis=1, center=20)
    fit_obj = fit_obj.remove_linear_term()
    fit = fit_obj.fit(max_order=2)
    assert np.sum(np.abs(fit['rel_err'])) > 0.0

@pytest.mark.rst
def test_load_exp(kamzik_converter: rst.KamzikConverter, input_file: rst.CXIStore):
    log_data = kamzik_converter.cxi_get(['basis_vectors', 'log_translations'])
    data = rst.STData(input_file, **log_data)
    data = data.load('data')
    for attr in ('y_pixel_size', 'translations', 'basis_vectors', 'x_pixel_size',
                 'good_frames', 'num_threads', 'input_file'):
        assert attr in data.contents()

@pytest.mark.rst
def test_kamzik_converter(kamzik_converter: rst.KamzikConverter):
    keys = kamzik_converter.cxi_keys()
    assert 'sim_translations' in keys and 'log_translations' in keys

@pytest.mark.rst
def test_defocus_sweep_exp(kamzik_converter: rst.KamzikConverter, input_file: rst.CXIStore, crop: rst.Crop,
                           good_frames_list: np.ndarray, defocus: float, distance: float, wavelength: float):
    log_data = kamzik_converter.cxi_get(['basis_vectors', 'log_translations'])
    data = rst.STData(input_file, **log_data, distance=distance, wavelength=wavelength, transform=crop)
    data = data.load('data').mask_frames(good_frames_list).update_mask(vmax=100000).integrate_data()

    defoci = np.linspace(0.5 * defocus, 2.0 * defocus, 50)
    sweep_scan = data.defocus_sweep(defoci_x=defoci, size=50)
    assert np.all(np.asarray(sweep_scan) > 0.0)

@pytest.mark.rst
def test_st_udpate_exp(kamzik_converter: rst.KamzikConverter, input_file: rst.CXIStore, crop: rst.Crop,
                       good_frames_list: np.ndarray, defocus: float, distance: float, wavelength: float):
    log_data = kamzik_converter.cxi_get(['basis_vectors', 'log_translations'])
    data = rst.STData(input_file, **log_data, distance=distance, wavelength=wavelength, transform=crop)
    data = data.load('data').mask_frames(good_frames_list).update_mask(vmax=100000).integrate_data()
    data = data.update_defocus(defocus)

    st_obj = data.get_st()
    h0 = st_obj.find_hopt()
    st_res = st_obj.train_adapt(search_window=(0.0, 10.0, 0.1), h0=h0, blur=8.0)
    data.import_st(st_res)
    fit = data.get_fit(axis=1).remove_linear_term().fit(max_order=3)
    assert np.sum(np.abs(fit['rel_err'])) > 0.0



from typing import Optional, Tuple
import numpy as np

def byte_align(array: np.ndarray, n: Optional[int]=None,
               dtype: Optional[np.dtype]=None) -> np.ndarray:
    """Function that takes a numpy array and checks it is aligned on an n-byte
    boundary, where `n` is an optional parameter. If `n` is not provided
    then this function will inspect the CPU to determine alignment. If the
    array is aligned then it is returned without further ado.  If it is not
    aligned then a new array is created and the data copied in, but aligned
    on the n-byte boundary.

    `dtype` is an optional argument that forces the resultant array to be
    of that dtype.
    """
    ...

def is_byte_aligned(array: np.ndarray, n: Optional[int]=None) -> bool:
    """Function that takes a numpy array and checks it is aligned on an n-byte
    boundary, where `n` is an optional parameter, returning `True` if it is,
    and `False` if it is not. If `n` is not provided then this function will
    inspect the CPU to determine alignment.
    """
    ...

def empty_aligned(shape: Tuple[int, ...], dtype: str='float64', order: str='C',
                  n: Optional[int]=None) -> np.ndarray:
    """Function that returns an empty numpy array that is n-byte aligned,
    where `n` is determined by inspecting the CPU if it is not
    provided.

    The alignment is given by the final optional argument, `n`. If
    `n` is not provided then this function will inspect the CPU to
    determine alignment. The rest of the arguments are as per
    :func:`numpy.empty`.
    """
    ...

def zeros_aligned(shape: Tuple[int, ...], dtype: str='float64', order: str='C',
                  n: Optional[int]=None) -> np.ndarray:
    """Function that returns a numpy array of zeros that is n-byte aligned,
    where `n` is determined by inspecting the CPU if it is not provided.

    The alignment is given by the final optional argument, `n`. If
    `n` is not provided then this function will inspect the CPU to
    determine alignment. The rest of the arguments are as per
    :func:`numpy.zeros`.
    """
    ...

def ones_aligned(shape: Tuple[int, ...], dtype: str='float64', order: str='C',
                 n: Optional[int]=None) -> np.ndarray:
    """Function that returns a numpy array of ones that is n-byte aligned,
    where `n` is determined by inspecting the CPU if it is not
    provided.

    The alignment is given by the final optional argument, `n`. If
    `n` is not provided then this function will inspect the CPU to
    determine alignment. The rest of the arguments are as per
    :func:`numpy.ones`.
    """
    ...

class FFTW:
    """FFTW is a class for computing a variety of discrete Fourier
    transforms of multidimensional, strided arrays using the FFTW
    library. The interface is designed to be somewhat pythonic, with
    the correct transform being inferred from the dtypes of the passed
    arrays.

    The exact scheme may be either directly specified with the
    `direction` parameter or inferred from the dtypes and relative
    shapes of the input arrays. Information on which shapes and dtypes
    imply which transformations is available in the :ref:`FFTW schemes
    <scheme_table>`. If a match is found, the plan corresponding to that
    scheme is created, operating on the arrays that are passed in. If no
    scheme can be created then a `ValueError` is raised.

    The actual transformation is performed by calling the
    :meth:`~pyfftw.FFTW.execute` method.

    The arrays can be updated by calling the
    :meth:`~pyfftw.FFTW.update_arrays` method.

    The created instance of the class is itself callable, and can perform
    the execution of the FFT, both with or without array updates, returning
    the result of the FFT. Unlike calling the :meth:`~pyfftw.FFTW.execute`
    method, calling the class instance will also optionally normalise the
    output as necessary. Additionally, calling with an input array update
    will also coerce that array to be the correct dtype.

    See the documentation on the :meth:`~pyfftw.FFTW.__call__` method
    for more information.
    """

    N                   : int
    simd_aligned        : bool
    input_alignment     : int
    output_alignment    : int
    flags               : Tuple[str, ...]
    input_array         : np.ndarray
    output_array        : np.ndarray
    input_strides       : Tuple[int, ...]
    output_strides      : Tuple[int, ...]
    input_shape         : Tuple[int, ...]
    output_shape        : Tuple[int, ...]
    input_dtype         : np.dtype
    output_dtype        : np.dtype
    direction           : str
    axes                : Tuple[int, ...]
    normalize_idft      : bool
    ortho               : bool

    def __init__(self, input_array: np.ndarray, output_array: np.ndarray,
                 axes: Tuple[int, ...]=(-1,), direction: str='FFTW_FORWARD',
                 flags: Tuple[str, ...]=('FFTW_MEASURE',), threads: int=1,
                 planning_timelimit: Optional[float]=None, normalise_idft: bool=True,
                 ortho: bool=False) -> None:
        r"""**Arguments**:

        * `input_array` and `output_array` should be numpy arrays.
          The contents of these arrays will be destroyed by the planning
          process during initialisation. Information on supported
          dtypes for the arrays is :ref:`given below <scheme_table>`.

        * `axes` describes along which axes the DFT should be taken.
          This should be a valid list of axes. Repeated axes are
          only transformed once. Invalid axes will raise an `IndexError`
          exception. This argument is equivalent to the same
          argument in :func:`numpy.fft.fftn`, except for the fact that
          the behaviour of repeated axes is different (`numpy.fft`
          will happily take the fft of the same axis if it is repeated
          in the `axes` argument). Rudimentary testing has suggested
          this is down to the underlying FFTW library and so unlikely
          to be fixed in these wrappers.

        * The `direction` parameter describes what sort of
          transformation the object should compute. This parameter is
          poorly named for historical reasons: older versions of pyFFTW
          only supported forward and backward transformations, for which
          this name made sense. Since then pyFFTW has been expanded to
          support real to real transforms as well and the name is not
          quite as descriptive.

          `direction` should either be a string, or, in the case of
          multiple real transforms, a list of strings. The two values
          corresponding to the DFT are

          * `'FFTW_FORWARD'`, which is the forward discrete Fourier
            transform, and
          * `'FFTW_BACKWARD'`, which is the backward discrete Fourier
            transform.

          Note that, for the two above options, only the Complex schemes
          allow a free choice for `direction`. The direction *must*
          agree with the the :ref:`table below <scheme_table>` if a Real
          scheme is used, otherwise a `ValueError` is raised.


          Alternatively, if you are interested in one of the real to real
          transforms, then pyFFTW supports four different discrete cosine
          transforms:

          * `'FFTW_REDFT00'`,
          * `'FFTW_REDFT01'`,
          * `'FFTW_REDFT10'`, and
          * `'FFTW_REDFT01'`,

          and four discrete sine transforms:

          * `'FFTW_RODFT00'`,
          * `'FFTW_RODFT01'`,
          * `'FFTW_RODFT10'`, and
          * `'FFTW_RODFT01'`.

          pyFFTW uses the same naming convention for these flags as FFTW:
          the `'REDFT'` part of the name is an acronym for 'real even
          discrete Fourier transform, and, similarly, `'RODFT'` stands
          for 'real odd discrete Fourier transform'. The trailing `'0'`
          is notation for even data (in terms of symmetry) and the
          trailing `'1'` is for odd data.

          Unlike the plain discrete Fourier transform, one may specify a
          different real to real transformation over each axis: for example,

          .. code-block:: none
             a = pyfftw.empty_aligned((128,128,128))
             b = pyfftw.empty_aligned((128,128,128))
             directions = ['FFTW_REDFT00', 'FFTW_RODFT11']
             transform = pyfftw.FFTW(a, b, axes=(0, 2), direction=directions)

          will create a transformation across the first and last axes
          with a discrete cosine transform over the first and a discrete
          sine transform over the last.

          Unfortunately, since this class is ultimately just a wrapper
          for various transforms implemented in FFTW, one cannot combine
          real transformations with real to complex transformations in a
          single object.

        .. _FFTW_flags:

        * `flags` is a list of strings and is a subset of the
          flags that FFTW allows for the planners:

          * `'FFTW_ESTIMATE'`, `'FFTW_MEASURE'`, `'FFTW_PATIENT'` and
            `'FFTW_EXHAUSTIVE'` are supported. These describe the
            increasing amount of effort spent during the planning
            stage to create the fastest possible transform.
            Usually `'FFTW_MEASURE'` is a good compromise. If no flag
            is passed, the default `'FFTW_MEASURE'` is used.
          * `'FFTW_UNALIGNED'` is supported.
            This tells FFTW not to assume anything about the
            alignment of the data and disabling any SIMD capability
            (see below).
          * `'FFTW_DESTROY_INPUT'` is supported.
            This tells FFTW that the input array can be destroyed during
            the transform, sometimes allowing a faster algorithm to be
            used. The default behaviour is, if possible, to preserve the
            input. In the case of the 1D Backwards Real transform, this
            may result in a performance hit. In the case of a backwards
            real transform for greater than one dimension, it is not
            possible to preserve the input, making this flag implicit
            in that case. A little more on this is given
            :ref:`below<scheme_table>`.
          * `'FFTW_WISDOM_ONLY'` is supported.
            This tells FFTW to raise an error if no plan for this transform
            and data type is already in the wisdom. It thus provides a method
            to determine whether planning would require additional effort or the
            cached wisdom can be used. This flag should be combined with the
            various planning-effort flags (`'FFTW_ESTIMATE'`,
            `'FFTW_MEASURE'`, etc.); if so, then an error will be raised if
            wisdom derived from that level of planning effort (or higher) is
            not present. If no planning-effort flag is used, the default of
            `'FFTW_ESTIMATE'` is assumed.
            Note that wisdom is specific to all the parameters, including the
            data alignment. That is, if wisdom was generated with input/output
            arrays with one specific alignment, using `'FFTW_WISDOM_ONLY'`
            to create a plan for arrays with any different alignment will
            cause the `'FFTW_WISDOM_ONLY'` planning to fail. Thus it is
            important to specifically control the data alignment to make the
            best use of `'FFTW_WISDOM_ONLY'`.

          The `FFTW planner flags documentation
          <http://www.fftw.org/fftw3_doc/Planner-Flags.html#Planner-Flags>`_
          has more information about the various flags and their impact.
          Note that only the flags documented here are supported.

        * `threads` tells the wrapper how many threads to use
          when invoking FFTW, with a default of 1. If the number
          of threads is greater than 1, then the GIL is released
          by necessity.

        * `planning_timelimit` is a floating point number that
          indicates to the underlying FFTW planner the maximum number of
          seconds it should spend planning the FFT. This is a rough
          estimate and corresponds to calling of `fftw_set_timelimit()`
          (or an equivalent dependent on type) in the underlying FFTW
          library. If `None` is set, the planner will run indefinitely
          until all the planning modes allowed by the flags have been
          tried. See the `FFTW planner flags page
          <http://www.fftw.org/fftw3_doc/Planner-Flags.html#Planner-Flags>`_
          for more information on this.

        .. _fftw_schemes:

        **Schemes**

        The currently supported full (so not discrete sine or discrete
        cosine) DFT schemes are as follows:

        .. _scheme_table:

        +----------------+-----------------------+------------------------+-----------+
        | Type           | `input_array.dtype` | `output_array.dtype` | Direction |
        +================+=======================+========================+===========+
        | Complex        | `complex64`         | `complex64`          | Both      |
        +----------------+-----------------------+------------------------+-----------+
        | Complex        | `complex128`        | `complex128`         | Both      |
        +----------------+-----------------------+------------------------+-----------+
        | Complex        | `clongdouble`       | `clongdouble`        | Both      |
        +----------------+-----------------------+------------------------+-----------+
        | Real           | `float32`           | `complex64`          | Forwards  |
        +----------------+-----------------------+------------------------+-----------+
        | Real           | `float64`           | `complex128`         | Forwards  |
        +----------------+-----------------------+------------------------+-----------+
        | Real           | `longdouble`        | `clongdouble`        | Forwards  |
        +----------------+-----------------------+------------------------+-----------+
        | Real\ :sup:`1` | `complex64`         | `float32`            | Backwards |
        +----------------+-----------------------+------------------------+-----------+
        | Real\ :sup:`1` | `complex128`        | `float64`            | Backwards |
        +----------------+-----------------------+------------------------+-----------+
        | Real\ :sup:`1` | `clongdouble`       | `longdouble`         | Backwards |
        +----------------+-----------------------+------------------------+-----------+

        \ :sup:`1`  Note that the Backwards Real transform for the case
        in which the dimensionality of the transform is greater than 1
        will destroy the input array. This is inherent to FFTW and the only
        general work-around for this is to copy the array prior to
        performing the transform. In the case where the dimensionality
        of the transform is 1, the default is to preserve the input array.
        This is different from the default in the underlying library, and
        some speed gain may be achieved by allowing the input array to
        be destroyed by passing the `'FFTW_DESTROY_INPUT'`
        :ref:`flag <FFTW_flags>`.

        The discrete sine and discrete cosine transforms are supported
        for all three real types.

        `clongdouble` typically maps directly to `complex256`
        or `complex192`, and `longdouble` to `float128` or
        `float96`, dependent on platform.

        The relative shapes of the arrays should be as follows:

        * For a Complex transform, `output_array.shape == input_array.shape`
        * For a Real transform in the Forwards direction, both the following
          should be true:

          * `output_array.shape[axes][-1] == input_array.shape[axes][-1]//2 + 1`
          * All the other axes should be equal in length.

        * For a Real transform in the Backwards direction, both the following
          should be true:

          * `input_array.shape[axes][-1] == output_array.shape[axes][-1]//2 + 1`
          * All the other axes should be equal in length.

        In the above expressions for the Real transform, the `axes`
        arguments denotes the unique set of axes on which we are taking
        the FFT, in the order passed. It is the last of these axes that
        is subject to the special case shown.

        The shapes for the real transforms corresponds to those
        stipulated by the FFTW library. Further information can be
        found in the FFTW documentation on the `real DFT
        <http://www.fftw.org/fftw3_doc/Guru-Real_002ddata-DFTs.html>`_.

        The actual arrangement in memory is arbitrary and the scheme
        can be planned for any set of strides on either the input
        or the output. The user should not have to worry about this
        and any valid numpy array should work just fine.

        What is calculated is exactly what FFTW calculates.
        Notably, this is an unnormalized transform so should
        be scaled as necessary (fft followed by ifft will scale
        the input by N, the product of the dimensions along which
        the DFT is taken). For further information, see the
        `FFTW documentation
        <http://www.fftw.org/fftw3_doc/What-FFTW-Really-Computes.html>`_.

        The FFTW library benefits greatly from the beginning of each
        DFT axes being aligned on the correct byte boundary, enabling
        SIMD instructions. By default, if the data begins on such a
        boundary, then FFTW will be allowed to try and enable
        SIMD instructions. This means that all future changes to
        the data arrays will be checked for similar alignment. SIMD
        instructions can be explicitly disabled by setting the
        FFTW_UNALIGNED flags, to allow for updates with unaligned
        data.

        :func:`~pyfftw.byte_align` and
        :func:`~pyfftw.empty_aligned` are two methods
        included with this module for producing aligned arrays.

        The optimum alignment for the running platform is provided
        by :data:`pyfftw.simd_alignment`, though a different alignment
        may still result in some performance improvement. For example,
        if the processor supports AVX (requiring 32-byte alignment) as
        well as SSE (requiring 16-byte alignment), then if the array
        is 16-byte aligned, SSE will still be used.

        It's worth noting that just being aligned may not be sufficient
        to create the fastest possible transform. For example, if the
        array is not contiguous (i.e. certain axes are displaced in
        memory), it may be faster to plan a transform for a contiguous
        array, and then rely on the array being copied in before the
        transform (which :class:`pyfftw.FFTW` will handle for you when
        accessed through :meth:`~pyfftw.FFTW.__call__`).
        """
        ...

    def __call__(self, input_array: Optional[np.ndarray]=None,
                 output_array: Optional[np.ndarray]=None,
                 normalise_idft: bool=None, ortho: bool=None) -> np.ndarray:
        """Calling the class instance (optionally) updates the arrays, then
        calls :meth:`~pyfftw.FFTW.execute`, before optionally normalising
        the output and returning the output array.

        It has some built-in helpers to make life simpler for the calling
        functions (as distinct from manually updating the arrays and
        calling :meth:`~pyfftw.FFTW.execute`).

        If `normalise_idft` is `True` (the default), then the output from
        an inverse DFT (i.e. when the direction flag is `'FFTW_BACKWARD'`) is
        scaled by 1/N, where N is the product of the lengths of input array on
        which the FFT is taken. If the direction is `'FFTW_FORWARD'`, this
        flag makes no difference to the output array.

        If `ortho` is `True`, then the output of both forward
        and inverse DFT operations is scaled by 1/sqrt(N), where N is the
        product of the lengths of input array on which the FFT is taken.  This
        ensures that the DFT is a unitary operation, meaning that it satisfies
        Parseval's theorem (the sum of the squared values of the transform
        output is equal to the sum of the squared values of the input).  In
        other words, the energy of the signal is preserved.

        If either `normalise_idft` or `ortho` are `True`, then
        ifft(fft(A)) = A.

        When `input_array` is something other than None, then the passed in
        array is coerced to be the same dtype as the input array used when the
        class was instantiated, the byte-alignment of the passed in array is
        made consistent with the expected byte-alignment and the striding is
        made consistent with the expected striding. All this may, but not
        necessarily, require a copy to be made.

        As noted in the :ref:`scheme table<scheme_table>`, if the FFTW
        instance describes a backwards real transform of more than one
        dimension, the contents of the input array will be destroyed. It is
        up to the calling function to make a copy if it is necessary to
        maintain the input array.

        `output_array` is always used as-is if possible. If the dtype, the
        alignment or the striding is incorrect for the FFTW object, then a
        `ValueError` is raised.

        The coerced input array and the output array (as appropriate) are
        then passed as arguments to
        :meth:`~pyfftw.FFTW.update_arrays`, after which
        :meth:`~pyfftw.FFTW.execute` is called, and then normalisation
        is applied to the output array if that is desired.

        Note that it is possible to pass some data structure that can be
        converted to an array, such as a list, so long as it fits the data
        requirements of the class instance, such as array shape.

        Other than the dtype and the alignment of the passed in arrays, the
        rest of the requirements on the arrays mandated by
        :meth:`~pyfftw.FFTW.update_arrays` are enforced.

        A `None` argument to either keyword means that that array is not
        updated.

        The result of the FFT is returned. This is the same array that is used
        internally and will be overwritten again on subsequent calls. If you
        need the data to persist longer than a subsequent call, you should
        copy the returned array.
        """
        ...

    def update_arrays(self, new_input_array: np.ndarray, new_output_array: np.ndarray) -> None:
        """Update the arrays upon which the DFT is taken.

        The new arrays should be of the same dtypes as the originals, the same
        shapes as the originals and should have the same strides between axes.
        If the original data was aligned so as to allow SIMD instructions
        (e.g. by being aligned on a 16-byte boundary), then the new array must
        also be aligned so as to allow SIMD instructions (assuming, of
        course, that the `FFTW_UNALIGNED` flag was not enabled).

        The byte alignment requirement extends to requiring natural
        alignment in the non-SIMD cases as well, but this is much less
        stringent as it simply means avoiding arrays shifted by, say,
        a single byte (which invariably takes some effort to
        achieve!).

        If all these conditions are not met, a `ValueError` will
        be raised and the data will *not* be updated (though the
        object will still be in a sane state).
        """
        ...

    def execute(self) -> None:
        """Execute the planned operation, taking the correct kind of FFT of
        the input array (i.e. :attr:`FFTW.input_array`), and putting the
        result in the output array (i.e. :attr:`FFTW.output_array`).
        """
        ...



"""Transforms are common image transformations. They can be chained together
using :class:`pyrost.ComposeTransforms`. You pass a :class:`pyrost.Transform`
instance to a data container :class:`pyrost.STData`. All transform classes
are inherited from the abstract :class:`pyrost.Transform` class.

:class:`pyrost.STData` contains all the necessary data for the Speckle
Tracking algorithm, and provides a suite of data processing tools to work
with the data.

Examples:
    Load all the necessary data using a :func:`pyrost.STData.load` function.

    >>> import pyrost as rst
    >>> inp_file = rst.CXIStore('data.cxi')
    >>> data = rst.STData(input_file=inp_file)
    >>> data = data.load()
"""
from __future__ import annotations
from typing import Any, Dict, Iterable, Iterator, List, Optional, Tuple, Union
from weakref import ref
from multiprocessing import cpu_count
from tqdm.auto import tqdm
import numpy as np
from .aberrations_fit import AberrationsFit
from .data_container import DataContainer, dict_to_object
from .cxi_protocol import CXIStore, Indices
from .rst_update import SpeckleTracking
from .bin import median, median_filter, fft_convolve, ct_integrate

class Transform():
    """Abstract transform class."""

    def index_array(self, ss_idxs: np.ndarray, fs_idxs: np.ndarray) -> Tuple[np.ndarray, np.ndarray]:
        raise NotImplementedError

    def __repr__(self) -> str:
        return self.state_dict().__repr__()

    def __str__(self) -> str:
        return self.state_dict().__str__()


    def forward(self, inp: np.ndarray) -> np.ndarray:
        """Return a transformed image.

        Args:
            inp : Input image.

        Returns:
            Transformed image.
        """
        ss_idxs, fs_idxs = np.indices(inp.shape[-2:])
        ss_idxs, fs_idxs = self.index_array(ss_idxs, fs_idxs)
        return inp[..., ss_idxs, fs_idxs]

    def state_dict(self) -> Dict[str, Any]:
        raise NotImplementedError

class Crop(Transform):
    """Crop transform. Crops a frame according to a region of interest.

    Attributes:
        roi : Region of interest. Comprised of four elements `[y_min, y_max,
            x_min, x_max]`.
    """
    def __init__(self, roi: Iterable[int]) -> None:
        """
        Args:
            roi : Region of interest. Comprised of four elements `[y_min, y_max,
                x_min, x_max]`.
        """
        self.roi = roi

    def index_array(self, ss_idxs: np.ndarray, fs_idxs: np.ndarray) -> Tuple[np.ndarray, np.ndarray]:
        """Filter the indices of a frame `(ss_idxs, fs_idxs)` according to 
        the cropping transform.

        Args:
            ss_idxs: Slow axis indices of a frame.
            fs_idxs: Fast axis indices of a frame.

        Returns:
            A tuple of filtered frame indices `(ss_idxs, fs_idxs)`.
        """
        if ss_idxs.shape[0] == 1:
            return (ss_idxs[:, self.roi[2]:self.roi[3]],
                    fs_idxs[:, self.roi[2]:self.roi[3]])

        if ss_idxs.shape[1] == 1:
            return (ss_idxs[self.roi[0]:self.roi[1], :],
                    fs_idxs[self.roi[0]:self.roi[1], :])

        return (ss_idxs[self.roi[0]:self.roi[1], self.roi[2]:self.roi[3]],
                fs_idxs[self.roi[0]:self.roi[1], self.roi[2]:self.roi[3]])

    def state_dict(self) -> Dict[str, Any]:
        """Returns the state of the transform as a dict.

        Returns:
            A dictionary with all the attributes.
        """
        return {'roi': self.roi[:]}

class Downscale(Transform):
    """Downscale the image by a integer ratio.

    Attributes:
        scale : Downscaling integer ratio.
    """
    def __init__(self, scale: int) -> None:
        """
        Args:
            scale : Downscaling integer ratio.
        """
        self.scale = scale

    def index_array(self, ss_idxs: np.ndarray, fs_idxs: np.ndarray) -> Tuple[np.ndarray, np.ndarray]:
        """Filter the indices of a frame `(ss_idxs, fs_idxs)` according to
        the downscaling transform.

        Args:
            ss_idxs: Slow axis indices of a frame.
            fs_idxs: Fast axis indices of a frame.

        Returns:
            A tuple of filtered frame indices `(ss_idxs, fs_idxs)`.
        """
        return (ss_idxs[::self.scale, ::self.scale], fs_idxs[::self.scale, ::self.scale])

    def state_dict(self) -> Dict[str, Any]:
        """Returns the state of the transform as a dict.

        Returns:
            A dictionary with all the attributes.
        """
        return {'scale': self.scale}

class Mirror(Transform):
    """Mirror the data around an axis.

    Attributes:
        axis : Axis of reflection.
    """
    def __init__(self, axis: int) -> None:
        """
        Args:
            axis : Axis of reflection.
        """
        if axis not in [0, 1]:
            raise ValueError('Axis must equal to 0 or 1')
        self.axis = axis

    def index_array(self, ss_idxs: np.ndarray, fs_idxs: np.ndarray) -> Tuple[np.ndarray, np.ndarray]:
        """Filter the indices of a frame `(ss_idxs, fs_idxs)` according to 
        the mirroring transform.

        Args:
            ss_idxs: Slow axis indices of a frame.
            fs_idxs: Fast axis indices of a frame.

        Returns:
            A tuple of filtered frame indices `(ss_idxs, fs_idxs)`.
        """
        if self.axis == 0:
            return (ss_idxs[::-1], fs_idxs[::-1])
        if self.axis == 1:
            return (ss_idxs[:, ::-1], fs_idxs[:, ::-1])
        raise ValueError('Axis must equal to 0 or 1')

    def state_dict(self) -> Dict[str, Any]:
        """Returns the state of the transform as a dict.

        Returns:
            A dictionary with all the attributes.
        """
        return {'axis': self.axis}

class ComposeTransforms(Transform):
    """Composes several transforms together.

    Attributes:
        transforms: List of transforms.
    """
    transforms : List[Transform]

    def __init__(self, transforms: List[Transform]) -> None:
        """
        Args:
            transforms: List of transforms.
        """
        self.transforms = []
        try:
            for transform in transforms:
                pdict = transform.state_dict()
                self.transforms.append(type(transform)(**pdict))
        except TypeError:
            raise TypeError('Invalid argument, must be a sequence of transforms.')
        else:
            if len(self.transforms) < 2:
                raise ValueError('Two or more transforms are needed to compose.')


    def __iter__(self) -> Iterator[Transform]:
        return self.transforms.__iter__()

    def __getitem__(self, idx: Indices) -> Transform:
        return self.transforms[idx]

    def index_array(self, ss_idxs: np.ndarray, fs_idxs: np.ndarray) -> Tuple[np.ndarray, np.ndarray]:
        """Filter the indices of a frame `(ss_idxs, fs_idxs)` according to 
        the composed transform.

        Args:
            ss_idxs: Slow axis indices of a frame.
            fs_idxs: Fast axis indices of a frame.

        Returns:
            A tuple of filtered frame indices `(ss_idxs, fs_idxs)`.
        """
        for transform in self:
            ss_idxs, fs_idxs = transform.index_array(ss_idxs, fs_idxs)
        return ss_idxs, fs_idxs

    def state_dict(self) -> Dict[str, Any]:
        """Returns the state of the transform as a dict.

        Returns:
            A dictionary with all the attributes.
        """
        return {'transforms': self.transforms[:]}

class STData(DataContainer):
    """Speckle tracking data container class. Needs a :class:`pyrost.CXIStore` file
    handler. Provides an interface to work with the data and contains a suite of
    tools for the R-PXST data processing pipeline. Also provides an interface to load
    from a file and save to a file any of the data attributes. The data frames can
    be tranformed using any of the :class:`pyrost.Transform` classes.

    Attributes:
        input_file : HDF5 or CXI file handler.
        transform : Frames transform object.

    Notes:
        **Necessary attributes**:

        * basis_vectors : Detector basis vectors
        * data : Measured intensity frames.
        * distance : Sample-to-detector distance [m].
        * frames : List of frame indices.
        * translations : Sample's translations [m].
        * wavelength : Incoming beam's wavelength [m].
        * x_pixel_size : Pixel's size along the horizontal detector axis [m].
        * y_pixel_size : Pixel's size along the vertical detector axis [m].

        **Optional attributes**:

        * defocus_x : Defocus distance for the horizontal detector axis [m].
        * defocus_y : Defocus distance for the vertical detector axis [m].
        * good_frames : An array of good frames' indices.
        * mask : Bad pixels mask.
        * num_threads : Number of threads used in computations.
        * output_file : Output file handler.
        * phase : Phase profile of lens' aberrations.
        * pixel_aberrations : Lens' aberrations along the horizontal and
          vertical axes in pixels.
        * pixel_translations : Sample's translations in the detector's
          plane in pixels.
        * reference_image : The unabberated reference image of the sample.
        * scale_map : Huber scale map.
        * whitefield : Measured frames' white-field.
        * whitefields : Set of dynamic white-fields for each of the measured
          images.
    """
    attr_set = {'input_file'}
    init_set = {'basis_vectors', 'data', 'distance', 'frames', 'translations', 'wavelength',
                'x_pixel_size', 'y_pixel_size', 'defocus_x', 'defocus_y', 'good_frames',
                'mask', 'num_threads', 'output_file', 'phase', 'pixel_aberrations',
                'pixel_translations', 'reference_image', 'scale_map', 'transform', 'whitefield',
                'whitefields'}

    # Necessary attributes
    input_file         : CXIStore
    transform           : Transform

    # Optional attributes
    basis_vectors       : Optional[np.ndarray]
    data                : Optional[np.ndarray]
    defocus_x           : Optional[float]
    defocus_y           : Optional[float]
    distance            : Optional[np.ndarray]
    frames              : Optional[np.ndarray]
    output_file         : Optional[CXIStore]
    phase               : Optional[np.ndarray]
    pixel_aberrations   : Optional[np.ndarray]
    reference_image     : Optional[np.ndarray]
    scale_map           : Optional[np.ndarray]
    translations        : Optional[np.ndarray]
    wavelength          : Union[float, np.ndarray, None]
    whitefields         : Optional[np.ndarray]
    x_pixel_size        : Union[float, np.ndarray, None]
    y_pixel_size        : Union[float, np.ndarray, None]

    # Automatially generated attributes
    good_frames         : Optional[np.ndarray]
    mask                : Optional[np.ndarray]
    num_threads         : Optional[int]
    pixel_translations  : Optional[np.ndarray]
    whitefield          : Optional[np.ndarray]

    def __init__(self, input_file: CXIStore, output_file: Optional[CXIStore]=None,
                 transform: Optional[Transform]=None, **kwargs: Union[int, float, np.ndarray]) -> None:
        """
        Args:
            input_file : HDF5 or CXI file handler of input files.
            output_file : Output file handler.
            transform : Frames transform object.
            kwargs : Dictionary of the necessary and optional data attributes specified
                in :class:`pyrost.STData` notes. All the necessary attributes must be
                provided

        Raises:
            ValueError : If any of the necessary attributes specified in :class:`pyrost.STData`
                notes have not been provided.
        """
        super(STData, self).__init__(input_file=input_file, output_file=output_file, 
                                     transform=transform, **kwargs)

        self._init_functions(num_threads=lambda: np.clip(1, 64, cpu_count()))
        if self.shape[0] > 0:
            self._init_functions(good_frames=lambda: np.arange(self.shape[0]))
        if self._isdata:
            self._init_functions(mask=lambda: np.ones(self.shape, dtype=bool))
            func = lambda: median(data=self.data[self.good_frames], axis=0,
                                  mask=self.mask[self.good_frames],
                                  num_threads=self.num_threads)
            self._init_functions(whitefield=func)
        if self._isdefocus:
            self._init_functions(defocus_y=lambda: self.get('defocus_x', None),
                                 pixel_translations=self._pixel_translations)

        self._init_attributes()

    @property
    def _isdata(self) -> bool:
        return self.data is not None

    @property
    def _isdefocus(self) -> bool:
        return self.defocus_x is not None

    @property
    def _isphase(self) -> bool:
        return not self.pixel_aberrations is None and not self.phase is None

    @property
    def shape(self) -> Tuple[int, int, int]:
        shape = [0, 0, 0]
        for attr, data in self.items():
            if attr in self.input_file.protocol and data is not None:
                kind = self.input_file.protocol.get_kind(attr)
                if kind == 'sequence':
                    shape[0] = data.shape[0]

        for attr, data in self.items():
            if attr in self.input_file.protocol and data is not None:
                kind = self.input_file.protocol.get_kind(attr)
                if kind == 'frame':
                    shape[1:] = data.shape

        for attr, data in self.items():
            if attr in self.input_file.protocol and data is not None:
                kind = self.input_file.protocol.get_kind(attr)
                if kind == 'stack':
                    shape[:] = data.shape
        return tuple(shape)

    def _pixel_translations(self) -> np.ndarray:
        pixel_translations = (self.translations[:, None] * self.basis_vectors).sum(axis=-1)
        mag = np.abs(self.distance / np.array([self.defocus_y, self.defocus_x]))
        pixel_translations *= mag / (self.basis_vectors**2).sum(axis=-1)
        pixel_translations -= pixel_translations[0]
        pixel_translations -= pixel_translations.mean(axis=0)
        return pixel_translations

    def pixel_map(self, dtype: np.dtype=np.float64) -> np.ndarray:
        """Return a preliminary pixel mapping.

        Args:
            dtype : The data type of the output pixel mapping.

        Returns:
            Pixel mapping array.
        """
        with self.input_file:
            self.input_file.update_indices()
            shape = self.input_file.read_shape()

        # Check if STData is integrated
        if self.shape[1] == 1:
            shape = (1, shape[1])
        if self.shape[2] == 1:
            shape = (shape[0], 1)

        ss_idxs, fs_idxs = np.indices(shape, dtype=dtype)
        if self.transform:
            ss_idxs, fs_idxs = self.transform.index_array(ss_idxs, fs_idxs)
        pixel_map = np.stack((ss_idxs, fs_idxs))

        if self._isdefocus:
            if self.defocus_y < 0.0:
                pixel_map = np.flip(pixel_map, axis=1)
            if self.defocus_x < 0.0:
                pixel_map = np.flip(pixel_map, axis=2)
        return np.asarray(pixel_map, order='C')

    @dict_to_object
    def load(self, attributes: Union[str, List[str], None]=None, idxs: Optional[Iterable[int]]=None,
             processes: int=1, verbose: bool=True) -> STData:
        """Load data attributes from the input files in `input_file` file handler object.

        Args:
            attributes : List of attributes to load. Loads all the data attributes
                contained in the file(s) by default.
            idxs : List of frame indices to load.
            processes : Number of parallel workers used during the loading.
            verbose : Set the verbosity of the loading process.

        Raises:
            ValueError : If attribute is not existing in the input file(s).
            ValueError : If attribute is invalid.

        Returns:
            New :class:`STData` object with the attributes loaded.
        """
        with self.input_file:
            self.input_file.update_indices()
            shape = self.input_file.read_shape()

            if attributes is None:
                attributes = [attr for attr in self.input_file.keys()
                              if attr in self.init_set]
            else:
                attributes = self.input_file.protocol.str_to_list(attributes)

            if idxs is None:
                idxs = self.input_file.indices()
            data_dict = {'frames': idxs, 'good_frames': None}

            for attr in attributes:
                if attr not in self.input_file.keys():
                    raise ValueError(f"No '{attr}' attribute in the input files")
                if attr not in self.init_set:
                    raise ValueError(f"Invalid attribute: '{attr}'")

                if self.transform and shape[0] * shape[1]:
                    ss_idxs, fs_idxs = np.indices(shape)
                    ss_idxs, fs_idxs = self.transform.index_array(ss_idxs, fs_idxs)
                    data = self.input_file.load_attribute(attr, idxs=idxs, ss_idxs=ss_idxs, fs_idxs=fs_idxs,
                                                          processes=processes, verbose=verbose)
                else:
                    data = self.input_file.load_attribute(attr, idxs=idxs, processes=processes,
                                                          verbose=verbose)

                data_dict[attr] = data

        return data_dict

    def save(self, attributes: Union[str, List[str], None]=None,
             mode: str='append', idxs: Optional[Iterable[int]]=None) -> None:
        """Save data arrays of the data attributes contained in the container to
        an output file.

        Args:
            attributes : List of attributes to save. Saves all the data attributes
                contained in the container by default.
            mode : Writing mode:

                * `append` : Append the data array to already existing dataset.
                * `insert` : Insert the data under the given indices `idxs`.
                * `overwrite` : Overwrite the existing dataset.

            idxs : A set of frame indices where the data is saved if `mode` is
                `insert`.

            verbose : Set the verbosity of the loading process.

        Raises:
            ValueError : If `output_file` is not defined inside the container.
        """
        if self.output_file is None:
            raise ValueError("'output_file' is not defined inside the container")

        if attributes is None:
            attributes = list(self.contents())

        with self.output_file:
            for attr in self.output_file.protocol.str_to_list(attributes):
                data = self.get(attr)
                if attr in self.output_file.protocol and data is not None:
                    kind = self.output_file.protocol.get_kind(attr)

                    if kind in ['stack', 'sequence']:
                        data = data[self.good_frames]

                    self.output_file.save_attribute(attr, np.asarray(data), mode=mode, idxs=idxs)

    @dict_to_object
    def clear(self, attributes: Union[str, List[str], None]=None) -> STData:
        """Clear the container.

        Args:
            attributes : List of attributes to clear in the container.

        Returns:
            New :class:`STData` object with the attributes cleared.
        """
        if attributes is None:
            attributes = self.keys()
        data_dict = {}
        for attr in self.input_file.protocol.str_to_list(attributes):
            data = self.get(attr)
            if attr in self and isinstance(data, np.ndarray):
                data_dict[attr] = None
        return data_dict

    @dict_to_object
    def update_output_file(self, output_file: CXIStore) -> STData:
        """Return a new :class:`STData` object with the new output
        file handler.

        Args:
            output_file : A new output file handler.

        Returns:
            New :class:`STData` object with the new output file
            handler.
        """
        return {'output_file': output_file}

    @dict_to_object
    def integrate_data(self, axis: int=0) -> STData:
        """Return a new :class:`STData` object with the `data` summed
        over the `axis`. Clear all the 2D and 3D data attributes inside the
        container.

        Args:
            axis : Axis along which a sum is performed.

        Returns:
            New :class:`STData` object with the stack of measured
            frames integrated along the given axis.
        """
        if self._isdata:
            data_dict = {}

            for attr, data in self.items():
                if attr in self.input_file.protocol and data is not None:
                    kind = self.input_file.protocol.get_kind(attr)
                    if kind in ['stack', 'frame']:
                        data_dict[attr] = None

            data_dict['data'] = (self.data * self.mask).sum(axis=axis - 2, keepdims=True)

            return data_dict

        raise AttributeError('data has not been loaded')

    @dict_to_object
    def mask_frames(self, good_frames: Optional[Iterable[int]]=None) -> STData:
        """Return a new :class:`STData` object with the updated good frames
        mask. Mask empty frames by default.

        Args:
            good_frames : List of good frames' indices. Masks empty frames
                if not provided.

        Returns:
            New :class:`STData` object with the updated `good_frames` and
            `whitefield`.
        """
        if good_frames is None:
            good_frames = np.where(self.data.sum(axis=(1, 2)) > 0)[0]
        return {'good_frames': np.asarray(good_frames), 'whitefield': None}

    @dict_to_object
    def update_mask(self, method: str='perc-bad', pmin: float=0., pmax: float=99.99,
                    vmin: int=0, vmax: int=65535, update: str='reset') -> STData:
        """Return a new :class:`STData` object with the updated bad pixels
        mask.

        Args:
            method : Bad pixels masking methods:

                * `no-bad` (default) : No bad pixels.
                * `range-bad` : Mask the pixels which values lie outside
                  of (`vmin`, `vmax`) range.
                * `perc-bad` : Mask the pixels which values lie outside
                  of the (`pmin`, `pmax`) percentiles.

            vmin : Lower intensity bound of 'range-bad' masking method.
            vmax : Upper intensity bound of 'range-bad' masking method.
            pmin : Lower percentage bound of 'perc-bad' masking method.
            pmax : Upper percentage bound of 'perc-bad' masking method.
            update : Multiply the new mask and the old one if `multiply`,
                use the new one if `reset`.

        Returns:
            New :class:`STData` object with the updated `mask`.
        """
        if update == 'reset':
            data = self.data
        elif update == 'multiply':
            data = self.data * self.mask
        else:
            raise ValueError(f'Invalid update keyword: {update:s}')

        if method == 'no-bad':
            mask = np.ones(self.shape, dtype=bool)
        elif method == 'range-bad':
            mask = (data >= vmin) & (data < vmax)
        elif method == 'perc-bad':
            average = median_filter(data, (1, 3, 3), num_threads=self.num_threads)
            offsets = (data.astype(np.int32) - average.astype(np.int32))
            mask = (offsets >= np.percentile(offsets, pmin)) & \
                   (offsets <= np.percentile(offsets, pmax))
        else:
            ValueError('invalid method argument')

        if update == 'reset':
            return {'mask': mask, 'whitefield': None}
        if update == 'multiply':
            return {'mask': mask * self.mask, 'whitefield': None}
        raise ValueError(f'Invalid update keyword: {update}')

    @dict_to_object
    def update_transform(self, transform: Transform) -> STData:
        """Return a new :class:`STData` object with the updated transform object.

        Args:
            transform : New :class:`Transform` object.

        Returns:
            New :class:`STData` object with the updated transform object.
        """
        data_dict = {'transform': transform}

        if self.transform is None:
            for attr, data in self.items():
                if attr in self.input_file.protocol and data is not None:
                    kind = self.input_file.protocol.get_kind(attr)
                    if kind in ['stack', 'frame']:
                        data = transform.forward(data)
                    data_dict[attr] = data

            return data_dict

        for attr, data in self.items():
            if attr in self.input_file.protocol and data is not None:
                kind = self.input_file.protocol.get_kind(attr)
                if kind in ['stack', 'frame']:
                    data_dict[attr] = None
        return data_dict

    @dict_to_object
    def update_whitefield(self) -> STData:
        """Return a new :class:`STData` object with the updated `whitefield`.

        Returns:
            New :class:`STData` object with the updated `whitefield`.
        """
        return {'whitefield': None}

    @dict_to_object
    def update_defocus(self, defocus_x: float, defocus_y: Optional[float]=None) -> STData:
        """Return a new :class:`STData` object with the updated defocus
        distances `defocus_x` and `defocus_y` for the horizontal and
        vertical detector axes accordingly. Update `pixel_translations`
        based on the new defocus distances.

        Args:
            defocus_x : Defocus distance for the horizontal detector axis [m].
            defocus_y : Defocus distance for the vertical detector axis [m].
                Equals to `defocus_x` if it's not provided.

        Returns:
            New :class:`STData` object with the updated `defocus_y`,
            `defocus_x`, and `pixel_translations`.
        """
        if defocus_y is None:
            defocus_y = defocus_x
        return {'defocus_y': defocus_y, 'defocus_x': defocus_x,
                'pixel_translations': None}

    def import_st(self, st_obj: SpeckleTracking) -> None:
        """Update `pixel_aberrations`, `phase`, `reference_image`, and `scale_map`
        based on the data from `st_obj` object. `st_obj` must be derived from this
        data container, an error is raised otherwise.

        Args:
            st_obj : :class:`SpeckleTracking` object derived from this
                data container.

        Raises:
            ValueError : If `st_obj` wasn't derived from this data container.
        """
        if st_obj.parent() is not self:
            raise ValueError("'st_obj' wasn't derived from this data container")
        # Update phase, pixel_aberrations, and reference_image
        dpm_y, dpm_x = (st_obj.pixel_map - self.pixel_map())
        dpm_y -= dpm_y.mean()
        dpm_x -= dpm_x.mean()
        self.pixel_aberrations = np.stack((dpm_y, dpm_x))

        # Calculate magnification for horizontal and vertical axes
        mag_y = np.abs((self.distance + self.defocus_y) / self.defocus_y)
        mag_x = np.abs((self.distance + self.defocus_x) / self.defocus_x)

        # Calculate the distance between the reference and the detector plane
        dist_y = self.distance * (1 - mag_y**-1)
        dist_x = self.distance * (1 - mag_x**-1)

        # dTheta = delta_pix / distance / magnification * du
        # Phase = 2 * pi / wavelength * Integrate[dTheta, delta_pix]
        phase = ct_integrate(sy_arr=self.y_pixel_size**2 / dist_y / mag_y * dpm_y,
                             sx_arr=self.x_pixel_size**2 / dist_x / mag_x * dpm_x)
        self.phase = 2.0 * np.pi / self.wavelength * phase
        self.reference_image = st_obj.reference_image
        self.scale_map = st_obj.scale_map

    def fit_phase(self, center: int=0, axis: int=1, max_order: int=2, xtol: float=1e-14,
                  ftol: float=1e-14, loss: str='cauchy') -> Dict[str, Union[float, np.ndarray]]:
        """Fit `pixel_aberrations` with the polynomial function using nonlinear
        least-squares algorithm. The function uses least-squares algorithm from
        :func:`scipy.optimize.least_squares`.

        Args:
            center : Index of the zerro scattering angle or direct beam pixel.
            axis : Axis along which `pixel_aberrations` is fitted.
            max_order : Maximum order of the polynomial model function.
            xtol : Tolerance for termination by the change of the independent
                variables.
            ftol : Tolerance for termination by the change of the cost function.
            loss : Determines the loss function. The following keyword values are
                allowed:

                * `linear` : ``rho(z) = z``. Gives a standard
                  least-squares problem.
                * `soft_l1` : ``rho(z) = 2 * ((1 + z)**0.5 - 1)``. The smooth
                  approximation of l1 (absolute value) loss. Usually a good
                  choice for robust least squares.
                * `huber` : ``rho(z) = z if z <= 1 else 2*z**0.5 - 1``. Works
                  similarly to 'soft_l1'.
                * `cauchy` (default) : ``rho(z) = ln(1 + z)``. Severely weakens
                  outliers influence, but may cause difficulties in optimization
                  process.
                * `arctan` : ``rho(z) = arctan(z)``. Limits a maximum loss on
                  a single residual, has properties similar to 'cauchy'.

        Returns:
            A dictionary with the model fit information. The following fields
            are contained:

            * `c_3` : Third order aberrations coefficient [rad / mrad^3].
            * `c_4` : Fourth order aberrations coefficient [rad / mrad^4].
            * `fit` : Array of the polynomial function coefficients of the
              pixel aberrations fit.
            * `ph_fit` : Array of the polynomial function coefficients of
              the phase aberrations fit.
            * `rel_err` : Vector of relative errors of the fit coefficients.
            * `r_sq` : ``R**2`` goodness of fit.

        See Also:
            :func:`pyrost.AberrationsFit.fit` : Full details of the aberrations
            fitting algorithm.
        """
        if not self._isphase:
            raise ValueError("'phase' is not defined inside the container.")
        return self.get_fit(center=center, axis=axis).fit(max_order=max_order,
                                                          xtol=xtol, ftol=ftol,
                                                          loss=loss)

    def defocus_sweep(self, defoci_x: np.ndarray, defoci_y: Optional[np.ndarray]=None, size: int=51,
                      hval: Optional[float]=None, extra_args: Dict[str, Union[float, bool, str]]={},
                      return_extra: bool=False, verbose: bool=True) -> Tuple[List[float], Dict[str, np.ndarray]]:
        r"""Calculate a set of reference images for each defocus in `defoci` and
        return an average R-characteristic of an image (the higher the value the
        sharper reference image is). The kernel bandwidth `hval` is automatically
        estimated by default. Return the intermediate results if `return_extra`
        is True.

        Args:
            defoci_x : Array of defocus distances along the horizontal detector axis [m].
            defoci_y : Array of defocus distances along the vertical detector axis [m].
            hval : Kernel bandwidth in pixels for the reference image update. Estimated
                with :func:`pyrost.SpeckleTracking.find_hopt` for an average defocus value
                if None.
            size : Local variance filter size in pixels.
            extra_args : Extra arguments parser to the :func:`STData.get_st` and
                :func:`SpeckleTracking.update_reference` methods. The following
                keyword values are allowed:

                * `ds_y` : Reference image sampling interval in pixels along the
                  horizontal axis. The default value is 1.0.
                * `ds_x` : Reference image sampling interval in pixels along the
                  vertical axis. The default value is 1.0.
                * `aberrations` : Add `pixel_aberrations` to `pixel_map` of
                  :class:`SpeckleTracking` object if it's True. The default value
                  is False.
                * `ff_correction` : Apply dynamic flatfield correction if it's True.
                  The default value is False.
                * `ref_method` : Choose the reference image update algorithm. The
                  following keyword values are allowed:

                  * `KerReg` : Kernel regression algorithm.
                  * `LOWESS` : Local weighted linear regression.

                  The default value is 'KerReg'.

            return_extra : Return a dictionary with the intermediate results if True.
            verbose : Set the verbosity of the process.

        Returns:
            A tuple of two items ('r_vals', 'extra'). The elements are as
            follows:

            * `r_vals` : Array of the average values of `reference_image` gradients
              squared.
            * `extra` : Dictionary with the intermediate results. Only if `return_extra`
              is True. Contains the following data:

              * reference_image : The generated set of reference profiles.
              * r_images : The set of local variance images of reference profiles.

        Notes:
            R-characteristic is called a local variance and is given by:

            .. math::
                R[i, j] = \frac{\sum_{i^{\prime} = -N / 2}^{N / 2}
                \sum_{j^{\prime} = -N / 2}^{N / 2} (I[i - i^{\prime}, j - j^{\prime}]
                - \bar{I}[i, j])^2}{\bar{I}^2[i, j]},

            where :math:`\bar{I}[i, j]` is a local mean and defined as follows:

            .. math::
                \bar{I}[i, j] = \frac{1}{N^2} \sum_{i^{\prime} = -N / 2}^{N / 2}
                \sum_{j^{\prime} = -N / 2}^{N / 2} I[i - i^{\prime}, j - j^{\prime}]

        See Also:
            :func:`pyrost.SpeckleTracking.update_reference` : reference image update
            algorithm.
        """
        if defoci_y is None:
            defoci_y = defoci_x.copy()

        ds_y = extra_args.get('ds_y', 1.0)
        ds_x = extra_args.get('ds_x', 1.0)
        aberrations = extra_args.get('aberrations', False)
        ff_correction = extra_args.get('ff_correction', False)
        ref_method = extra_args.get('ref_method', 'KerReg')

        r_vals = []
        extra = {'reference_image': [], 'r_image': []}
        kernel = np.ones(int(size)) / size
        df0_x, df0_y = defoci_x.mean(), defoci_y.mean()
        st_obj = self.update_defocus(df0_x, df0_y).get_st(ds_y=ds_y, ds_x=ds_x,
                                                          aberrations=aberrations,
                                                          ff_correction=ff_correction)
        if hval is None:
            hval = st_obj.find_hopt(method=ref_method)

        for df1_x, df1_y in tqdm(zip(defoci_x.ravel(), defoci_y.ravel()),
                               total=defoci_x.size, disable=not verbose,
                               desc='Generating defocus sweep'):
            st_obj.di_pix *= np.abs(df0_y / df1_y)
            st_obj.dj_pix *= np.abs(df0_x / df1_x)
            df0_x, df0_y = df1_x, df1_y
            st_obj.update_reference.inplace_update(hval=hval, method=ref_method)
            extra['reference_image'].append(st_obj.reference_image)
            mean = st_obj.reference_image.copy()
            mean_sq = st_obj.reference_image**2
            if st_obj.reference_image.shape[0] > size:
                mean = fft_convolve(mean, kernel, mode='reflect', axis=0,
                                    num_threads=self.num_threads)[size // 2:-size // 2]
                mean_sq = fft_convolve(mean_sq, kernel, mode='reflect', axis=0,
                                       num_threads=self.num_threads)[size // 2:-size // 2]
            if st_obj.reference_image.shape[1] > size:
                mean = fft_convolve(mean, kernel, mode='reflect', axis=1,
                                    num_threads=self.num_threads)[:, size // 2:-size // 2]
                mean_sq = fft_convolve(mean_sq, kernel, mode='reflect', axis=1,
                                       num_threads=self.num_threads)[:, size // 2:-size // 2]
            r_image = (mean_sq - mean**2) / mean**2
            extra['r_image'].append(r_image)
            r_vals.append(np.mean(r_image))

        if return_extra:
            return r_vals, extra
        return r_vals

    def get_st(self, ds_y: float=1.0, ds_x: float=1.0, aberrations: bool=False,
               ff_correction: bool=False) -> SpeckleTracking:
        """Return :class:`SpeckleTracking` object derived from the container.
        Return None if `defocus_x` or `defocus_y` doesn't exist in the container.

        Args:
            ds_y : Reference image sampling interval in pixels along the vertical
                axis.
            ds_x : Reference image sampling interval in pixels along the
                horizontal axis.
            aberrations : Add `pixel_aberrations` to `pixel_map` of
                :class:`SpeckleTracking` object if it's True.
            ff_correction : Apply dynamic flatfield correction if it's True.

        Returns:
            An instance of :class:`SpeckleTracking` derived from the container.
            None if `defocus_x` or `defocus_y` are not defined.
        """
        if not self._isdefocus:
            raise ValueError("'defocus_x' is not defined inside the container.")

        if np.issubdtype(self.data.dtype, np.uint32):
            dtypes = SpeckleTracking.dtypes_32
        else:
            dtypes = SpeckleTracking.dtypes_64

        data = np.asarray((self.mask * self.data)[self.good_frames],
                          order='C', dtype=dtypes['data'])
        whitefield = np.asarray(self.whitefield, order='C', dtype=dtypes['whitefield'])
        dij_pix = np.asarray(np.swapaxes(self.pixel_translations[self.good_frames], 0, 1),
                             order='C', dtype=dtypes['dij_pix'])

        if ff_correction and self.whitefields is not None:
            np.rint(data * np.where(self.whitefields > 0, whitefield / self.whitefields, 1.),
                    out=data, casting='unsafe')

        pixel_map = self.pixel_map(dtype=dtypes['pixel_map'])

        if aberrations:
            pixel_map += self.pixel_aberrations
            if self.scale_map is None:
                scale_map = None
            else:
                scale_map = np.asarray(self.scale_map, order='C', dtype=dtypes['scale_map'])
            return SpeckleTracking(parent=ref(self), data=data, dj_pix=dij_pix[1],
                                   di_pix=dij_pix[0], num_threads=self.num_threads,
                                   pixel_map=pixel_map, scale_map=scale_map, ds_y=ds_y,
                                   ds_x=ds_x, whitefield=whitefield)

        return SpeckleTracking(parent=ref(self), data=data, dj_pix=dij_pix[1], di_pix=dij_pix[0],
                               num_threads=self.num_threads, pixel_map=pixel_map, ds_y=ds_y,
                               ds_x=ds_x, whitefield=whitefield)

    def get_fit(self, center: int=0, axis: int=1) -> AberrationsFit:
        """Return an :class:`AberrationsFit` object for parametric regression
        of the lens' aberrations profile. Raises an error if 'defocus_x' or
        'defocus_y' is not defined.

        Args:
            center : Index of the zerro scattering angle or direct beam pixel.
            axis : Detector axis along which the fitting is performed.

        Raises:
            ValueError : If 'defocus_x' or 'defocus_y' is not defined in the
                container.

        Returns:
            An instance of :class:`AberrationsFit` class.
        """
        if not self._isphase:
            raise ValueError("'phase' or 'pixel_aberrations' are not defined inside the container.")

        data_dict = {attr: self.get(attr) for attr in AberrationsFit.attr_set if attr in self}
        if axis == 0:
            data_dict.update({attr: self.get(data_attr)
                              for attr, data_attr in AberrationsFit.y_lookup.items()})
        elif axis == 1:
            data_dict.update({attr: self.get(data_attr)
                              for attr, data_attr in AberrationsFit.x_lookup.items()})
        else:
            raise ValueError(f'invalid axis value: {axis:d}')

        data_dict['defocus'] = np.abs(data_dict['defocus'])
        if center <= self.shape[axis - 2]:
            data_dict['pixels'] = np.arange(self.shape[axis - 2]) - center
            data_dict['pixel_aberrations'] = data_dict['pixel_aberrations'][axis].mean(axis=1 - axis)
        elif center >= self.shape[axis - 2] - 1:
            data_dict['pixels'] = center - np.arange(self.shape[axis - 2])
            idxs = np.argsort(data_dict['pixels'])
            data_dict['pixel_aberrations'] = -data_dict['pixel_aberrations'][axis].mean(axis=1 - axis)[idxs]
            data_dict['pixels'] = data_dict['pixels'][idxs]
        else:
            raise ValueError('Origin must be outside of the region of interest')

        return AberrationsFit(parent=ref(self), **data_dict)

    def get_pca(self) -> Tuple[np.ndarray, np.ndarray, np.ndarray]:
        """Perform the Principal Component Analysis [PCA]_ of the measured data and
        return a set of eigen flatfields (EFF).

        Returns:
            A tuple of ('cor_data', 'effs', 'eig_vals'). The elements are
            as follows:

            * `cor_data` : Background corrected stack of measured frames.
            * `effs` : Set of eigen flat-fields.
            * `eig_vals` : Corresponding eigen values for each of the eigen
              flat-fields.

        References:
            .. [PCA] Vincent Van Nieuwenhove, Jan De Beenhouwer, Francesco De Carlo,
                    Lucia Mancini, Federica Marone, and Jan Sijbers, "Dynamic intensity
                    normalization using eigen flat fields in X-ray imaging," Opt.
                    Express 23, 27975-27989 (2015).
        """
        if self._isdata:

            dtype = np.promote_types(self.whitefield.dtype, int)
            cor_data = np.zeros(self.shape, dtype=dtype)[self.good_frames]
            np.subtract(self.data[self.good_frames], self.whitefield, dtype=dtype,
                        where=self.mask[self.good_frames], out=cor_data)
            mat_svd = np.tensordot(cor_data, cor_data, axes=((1, 2), (1, 2)))
            eig_vals, eig_vecs = np.linalg.eig(mat_svd)
            effs = np.tensordot(eig_vecs, cor_data, axes=((0,), (0,)))
            return cor_data, effs, eig_vals / eig_vals.sum()

        raise AttributeError('Data has not been loaded')

    @dict_to_object
    def update_whitefields(self, method: str='median', size: int=11,
                           cor_data: Optional[np.ndarray]=None,
                           effs: Optional[np.ndarray]=None) -> STData:
        """Return a new :class:`STData` object with a new set of dynamic whitefields.
        A set of whitefields are generated by the dint of median filtering or Principal
        Component Analysis [PCA]_.

        Args:
            method : Method to generate a set of dynamic white-fields. The following
                keyword values are allowed:

                * `median` : Median `data` along the first axis.
                * `pca` : Generate a set of dynamic white-fields based on eigen flatfields
                  `effs` from the PCA. `effs` can be obtained with :func:`STData.get_pca`
                  method.

            size : Size of the filter window in pixels used for the 'median' generation
                method.
            cor_data : Background corrected stack of measured frames.
            effs : Set of Eigen flatfields used for the 'pca' generation method.

        Raises:
            ValueError : If the `method` keyword is invalid.
            AttributeError : If the `whitefield` is absent in the :class:`STData`
                container when using the 'pca' generation method.
            ValueError : If `effs` were not provided when using the 'pca' generation
                method.

        Returns:
            New :class:`STData` object with the updated `whitefields`.

        See Also:
            :func:`pyrost.STData.get_pca` : Method to generate eigen flatfields.
        """
        if self._isdata:

            if method == 'median':
                offsets = np.abs(self.data - self.whitefield)
                outliers = offsets < 3 * np.sqrt(self.whitefield)
                whitefields = median_filter(self.data, size=(size, 1, 1), mask=outliers,
                                            num_threads=self.num_threads)
            elif method == 'pca':
                if cor_data is None:
                    dtype = np.promote_types(self.whitefield.dtype, int)
                    cor_data = np.zeros(self.shape, dtype=dtype)
                    np.subtract(self.data, self.whitefield, dtype=dtype,
                                where=self.mask, out=cor_data)
                if effs is None:
                    raise ValueError('No eigen flat fields were provided')

                weights = np.tensordot(cor_data, effs, axes=((1, 2), (1, 2)))
                weights /= np.sum(effs * effs, axis=(1, 2))
                whitefields = np.tensordot(weights, effs, axes=((1,), (0,)))
                whitefields += self.whitefield
            else:
                raise ValueError('Invalid method argument')

            return {'whitefields': whitefields}

        raise ValueError('Data has not been loaded')



"""pyrost.simulation (st_sim) provides wavefront propagation tools to simulate
speckle tracking scans. Wavefront propagation is based on the Fresnel diffraction
theory.

(c) Nikolay Ivanov, 2020.
"""
from .st_parameters import STParams
from .st_sim import STSim, STConverter



"""
Examples:

    :func:`pyrost.multislice.MSParams.import_default` generates the multislice
    experimental parameters, which could be later parsed to
    :class:`pyrost.multislice.MSPropagator` in order to perform the simulation.

    >>> import pyrost.multislice as ms_sim
    >>> ms_params = ms_sim.MSParams.import_default()
    >>> print(ms_params)
    {'multislice': {'x_max': 30.0, 'x_min': 0.0, 'x_step': 0.0001, '...': '...'},
     'material1': {'formula': 'W', 'density': 18.0}, 'material2': {'formula': 'SiC',
     'density': 2.8}, 'mll': {'focus': 1500.0, 'n_max': 8000, 'n_min': 100, '...': '...'}}
"""
from __future__ import annotations
import os
import re
from typing import Dict, List, Optional, Union
import numpy as np
from ..ini_parser import INIParser, ROOT_PATH
from ..bin import next_fast_len

MS_PARAMETERS = os.path.join(ROOT_PATH, 'config/ms_parameters.ini')

ELEMENTS = ('None', 'H', 'He', 'Li', 'Be', 'B', 'C', 'N', 'O', 'F', 'Ne',
            'Na', 'Mg', 'Al', 'Si', 'P', 'S', 'Cl', 'Ar', 'K', 'Ca', 'Sc', 'Ti', 'V',
            'Cr', 'Mn', 'Fe', 'Co', 'Ni', 'Cu', 'Zn', 'Ga', 'Ge', 'As', 'Se', 'Br',
            'Kr', 'Rb', 'Sr', 'Y', 'Zr', 'Nb', 'Mo', 'Tc', 'Ru', 'Rh', 'Pd', 'Ag',
            'Cd', 'In', 'Sn', 'Sb', 'Te', 'I', 'Xe', 'Cs', 'Ba', 'La', 'Ce', 'Pr',
            'Nd', 'Pm', 'Sm', 'Eu', 'Gd', 'Tb', 'Dy', 'Ho', 'Er', 'Tm', 'Yb', 'Lu',
            'Hf', 'Ta', 'W', 'Re', 'Os', 'Ir', 'Pt', 'Au', 'Hg', 'Tl', 'Pb', 'Bi',
            'Po', 'At', 'Rn', 'Fr', 'Ra', 'Ac', 'Th', 'Pa', 'U')

class BasicElement:
    """Bacis chemical element class.

    Attributes:
        name : Name of the element.
        mass : Atomic mass of the element.
        density : Atomic density of the element.
    """
    en_to_wl = 12398.419297617678 # h * c / e [eV * A]
    ref_cf = 2.7008646837561236e-6 # Avogadro * r_el / 2 / pi [A]

    def __init__(self, name: str, mass: float, density: float) -> None:
        """
        Args:
            name : Name of the element.
            mass : Atomic mass of the element.
            density : Atomic density of the element.
        """
        self.name, self.mass, self.density = name, mass, density

    def __repr__(self) -> str:
        out = {attr: self.__getattribute__(attr) for attr in ['name', 'density', 'mass']}
        return out.__repr__()

    def __str__(self) -> str:
        out = {attr: self.__getattribute__(attr) for attr in ['name', 'density', 'mass']}
        return out.__str__()

    def export_dict(self) -> Dict[str, Union[str, float]]:
        """Export object to a dictionary.

        Returns:
            Dictionary with element's name, mass, and density.
        """
        return {'formula': self.name, 'mass': self.mass, 'density': self.density}

    def get_ref_index(self, energy: Union[float, np.ndarray]) -> Union[float, np.ndarray]:
        r"""Calculates the complex refractive index for the given photon `energy`.

        Args:
            energy : Photon energy [eV].

        Returns:
            Linear absorption coefficient [cm^-1]

        Notes:
            The complex refractive index is customary denoted as:

            .. math::

                n = 1 - \delta + i \beta

            The real and imaginary components, :math:`\delta` and :math:`\beta` are
            given by [GISAXS]_:

            .. math::

                \delta = \frac{\rho N_a r_e \lambda f_1}{2 \pi m_a}

            .. math::

                \beta = \frac{\rho N_a r_e \lambda f_2}{2 \pi m_a}

            where :math:`\rho` is physical density, :math:`N_a` is Avogadro constant, :math:`m_a` is
            atomic molar mass, :math:`r_e` is radius of electron, :math:`\lambda` is wavelength,
            :math:`f_1` and :math:`f_2` are real and imaginary components of scattering factor.

        Reference:
            .. [GISAXS] http://gisaxs.com/index.php/Refractive_index
        """
        wavelength = self.en_to_wl / energy
        ref_idx = self.ref_cf * wavelength**2 * self.density * self.get_sf(energy) / self.mass
        return ref_idx

    def get_absorption_coefficient(self, energy: Union[float, np.ndarray]) -> Union[float, np.ndarray]:
        r"""Calculates the linear absorption coefficientfor the given photon `energy`.

        Args:
            energy : Photon energy [eV].

        Returns:
            Linear absorption coefficient [cm^-1]

        Notes:
            The absorption coefficient is the inverse of the absorption length and
            is given by [GISAXS]_:

            .. math::

                \mu = \frac{\rho N_a}{m_a} 2 r_e \lambda f_2 = \frac{4 \pi \beta}{\lambda}

            where :math:`\rho` is physical density, :math:`N_a` is Avogadro constant, :math:`m_a` is
            atomic molar mass, :math:`r_e` is radius of electron, :math:`\lambda` is wavelength,
            :math:`f_1` and :math:`f_2` are real and imaginary components of scattering factor.
        """
        wavelength = self.en_to_wl / energy
        return 4 * np.pi * self.get_ref_index(energy).imag / wavelength

class Element(BasicElement):
    """This class serves for accessing the scattering factors f1 and f2
    and atomic scattering factor of a chemical element `elem`.

    Attributes:
        name : Name of the chemical element `elem`.
        atom_num : Atomic number of `elem`.
        dbase : Database of scattering factors.
        asf_coeffs : Coefficients of atomic scattering factor.
        sf_coeffs : Scattering factors (`energy`, `f1`, `f2`).
        mass : Atomic mass [u].
        radius : Atomic radius [Angstrom].
        density : Density [g / cm^3].
    """
    dbase_lookup = {'Henke': 'data/henke_f1f2.npz', 'Chantler': 'data/chantler_f1f2.npz',
                    'BrCo': 'data/brco_f1f2.npz'}
    asf_dbase = 'data/henke_f0.npz'
    atom_dbase = 'data/atom_data.npz'

    def __init__(self, elem: Union[str, int], dbase: str='Chantler') -> None:
        """
        Args:
            elem : The element can be specified by its name (case sensitive) or its
                atomic number.
            dbase : Database of the tabulated scattering factors of each element. The
                following keywords are allowed:

                * `Henke` : (10 eV < E < 30 keV) [Henke]_.
                * `Chantler` : (11 eV < E < 405 keV) [Chantler]_.
                * `BrCo` : (30 eV < E < 509 keV) [BrCo]_.

        References:
            .. [Henke] http://henke.lbl.gov/optical_constants/asf.html
                       B.L. Henke, E.M. Gullikson, and J.C. Davis, *X-ray interactions:
                       photoabsorption, scattering, transmission, and reflection at
                       E=50-30000 eV, Z=1-92*, Atomic Data and Nuclear Data Tables
                       **54** (no.2) (1993) 181-342.
            .. [Chantler] http://physics.nist.gov/PhysRefData/FFast/Text/cover.html
                          http://physics.nist.gov/PhysRefData/FFast/html/form.html
                          C. T. Chantler, *Theoretical Form Factor, Attenuation, and
                          Scattering Tabulation for Z = 1 - 92 from E = 1 - 10 eV to E = 0.4 -
                          1.0 MeV*, J. Phys. Chem. Ref. Data **24** (1995) 71-643.
            .. [BrCo] http://www.bmsc.washington.edu/scatter/periodic-table.html
                      ftp://ftpa.aps.anl.gov/pub/cross-section_codes/
                      S. Brennan and P.L. Cowan, *A suite of programs for calculating
                      x-ray absorption, reflection and diffraction performance for a
                      variety of materials at arbitrary wavelengths*, Rev. Sci. Instrum.
                      **63** (1992) 850-853.
        """
        if isinstance(elem, str):
            name = elem
            self.atom_num = ELEMENTS.index(elem)
        elif isinstance(elem, int):
            name = ELEMENTS[elem]
            self.atom_num = elem
        else:
            raise ValueError('Wrong element: {:s}'.format(str(elem)))
        if dbase in self.dbase_lookup:
            self.dbase = dbase
        else:
            raise ValueError('Wrong database: {:s}'.format(dbase))

        with np.load(os.path.join(ROOT_PATH, self.atom_dbase)) as dbase:
            mass = dbase['mass'][self.atom_num - 1]
            density = dbase['density'][self.atom_num - 1]
            self.radius = dbase['radius'][self.atom_num - 1]

        super(Element, self).__init__(name, mass, density)

        with np.load(os.path.join(ROOT_PATH, self.asf_dbase)) as dbase:
            self.asf_coeffs = dbase[self.name]
        with np.load(os.path.join(ROOT_PATH, self.dbase_lookup[self.dbase])) as dbase:
            self.sf_coeffs = np.stack((dbase[self.name + '_E'], dbase[self.name + '_f1'],
                                       dbase[self.name + '_f2']))

    def get_asf(self, scat_vec: float) -> float:
        """Calculate atomic scattering factor for the given magnitude of
        scattering vector `scat_vec`.

        Args:
            scat_vec : Scattering vector magnitude [Angstrom^-1].

        Returns:
            Atomic scattering factor.
        """
        q_ofpi = scat_vec / 4 / np.pi
        asf = self.asf_coeffs[5] + sum(a * np.exp(-b * q_ofpi**2)
                                       for a, b in zip(self.asf_coeffs[:5], self.asf_coeffs[6:]))
        return asf

    def get_sf(self, energy: Union[float, np.ndarray]) -> Union[float, np.ndarray]:
        """Return a complex scattering factor (`f1` + 1j * `f2`) for the given photon
        `energy`.

        Args:
            energy : Photon energy [eV].

        Returns:
            Complex scattering factor.
        """
        if np.any(energy < self.sf_coeffs[0, 0]) or np.any(energy > self.sf_coeffs[0, -1]):
            exc_txt = 'Energy is out of bounds: ({0:.2f} - {1:.2f})'.format(self.sf_coeffs[0, 0],
                                                                            self.sf_coeffs[0, -1])
            raise ValueError(exc_txt)
        f_one = np.interp(energy, self.sf_coeffs[0], self.sf_coeffs[1]) + self.atom_num
        f_two = np.interp(energy, self.sf_coeffs[0], self.sf_coeffs[2])
        return f_one + 1j * f_two

class Material(BasicElement):
    """
    :class:`Material` serves for getting refractive index and absorption
    coefficient of a material specified by its chemical formula and density.

    Attributes:
        name : Name of the compound.
        elements : List of elements in the chemical formula.
        quantities : Coefficients in the chemical formula.
        mass : Molar mass [u].
        density : Atomic density of the element.
    """
    FORMULA_MATCHER = '(' + '|'.join(ELEMENTS[1:]) + ')'
    NUM_MATCHER = r'\d+'

    elements    : List[Element]
    quantities  : List[int]

    def __init__(self, formula: str, density: float, dbase: str='Chantler') -> None:
        """
        Args:
            elem : The element can be specified by its name (case sensitive) or its
                atomic number.
            density : Atomic density of the compound.
            dbase : Database of the tabulated scattering factors of each element. The
                following keywords are allowed:

                * `Henke` : (10 eV < E < 30 keV) [Henke]_.
                * `Chantler` : (11 eV < E < 405 keV) [Chantler]_.
                * `BrCo` : (30 eV < E < 509 keV) [BrCo]_.

        References:
            .. [Henke] http://henke.lbl.gov/optical_constants/asf.html
                       B.L. Henke, E.M. Gullikson, and J.C. Davis, *X-ray interactions:
                       photoabsorption, scattering, transmission, and reflection at
                       E=50-30000 eV, Z=1-92*, Atomic Data and Nuclear Data Tables
                       **54** (no.2) (1993) 181-342.
            .. [Chantler] http://physics.nist.gov/PhysRefData/FFast/Text/cover.html
                          http://physics.nist.gov/PhysRefData/FFast/html/form.html
                          C. T. Chantler, *Theoretical Form Factor, Attenuation, and
                          Scattering Tabulation for Z = 1 - 92 from E = 1 - 10 eV to E = 0.4 -
                          1.0 MeV*, J. Phys. Chem. Ref. Data **24** (1995) 71-643.
            .. [BrCo] http://www.bmsc.washington.edu/scatter/periodic-table.html
                      ftp://ftpa.aps.anl.gov/pub/cross-section_codes/
                      S. Brennan and P.L. Cowan, *A suite of programs for calculating
                      x-ray absorption, reflection and diffraction performance for a
                      variety of materials at arbitrary wavelengths*, Rev. Sci. Instrum.
                      **63** (1992) 850-853.
        """
        self.elements = []
        self.quantities = []
        mass = 0

        for m in re.finditer(self.FORMULA_MATCHER, formula):
            new_elem = Element(elem=m.group(0), dbase=dbase)
            m_num = re.match(self.NUM_MATCHER, formula[m.end():])
            quant = int(m_num.group(0)) if m_num else 1
            mass += new_elem.mass * quant
            self.elements.append(new_elem)
            self.quantities.append(quant)

        super(Material, self).__init__(formula, mass, density)

    def get_sf(self, energy: Union[float, np.ndarray]) -> Union[float, np.ndarray]:
        """Return a complex scattering factor (`f1` + 1j * `f2`) for the given photon
        `energy`.

        Args:
            energy : Photon energy [eV].

        Returns:
            Complex scattering factor.
        """
        return np.sum([elem.get_sf(energy) * quant
                       for elem, quant in zip(self.elements, self.quantities)], axis=0)

class MSParams(INIParser):
    """Container with the experimental parameters of
    one-dimensional multislice beam propagation simulation.
    All the experimental parameters are enlisted in :ref:`ms-parameters`.

    Attributes:
        x_min : Wavefront lower bound along the x axis [um].
        x_max : Wavefront upper bound along the x axis [um].
        x_step : Beam sample interval along the x axis [um].
        z_step : Distance between the slices [um].
        wl : Beam's wavelength [um].
        focus : MLL's focal distance [um].
        n_min : Zone number of the first layer in the MLL.
        n_max : Zone number of the last layer in the MLL.
        mll_sigma : MLL's bilayer interdiffusion length [um].
        mll_depth : MLL's thickness [um].
        mll_wl : Wavelength oh the MLL [um].
        mll_mat1 : The first MLL material.
        mll_mat2 : The second MLL material.

    See Also:
        :ref:`ms-parameters` : Full list of experimental parameters.
    """
    attr_dict = {'multislice':  ('x_max', 'x_min', 'x_step', 'z_step', 'wl'),
                 'material1':   ('formula', 'density'),
                 'material2':   ('formula', 'density'),
                 'mll':         ('focus', 'n_max', 'n_min', 'mll_sigma', 'mll_depth', 'mll_wl')}

    fmt_dict = {'multislice': 'float', 'material1/formula': 'str', 'material1/density': 'float',
                'material2/formula': 'str', 'material2/density': 'float', 'mll/n_min': 'int',
                'mll/n_max': 'int', 'mll': 'float'}

    # multislice attributes
    x_min       : float
    x_max       : float
    x_step      : float
    z_step      : float
    wl          : float

    # mll attributes
    n_min       : int
    n_max       : int
    focus       : float
    mll_sigma   : float
    mll_depth   : float
    mll_wl      : float

    def __init__(self, multislice: Dict[str, float], mll_mat1: Material,
                 mll_mat2: Material, mll: Dict[str, Union[int, float]]) -> None:
        """
        Args:
            multislice : A dictionary of multislice simulation parameters. The following
                elements are accepted:

                * `x_min`, `x_max` : Wavefront span along the x axis [um].
                * `x_step` : Beam sampling interval along the x axis [um].
                * `z_step` : Distance between the slices [um].
                * `wl` : Beam's wavelength [um].

            mll_mat1 : A dictionary of the first MLL material. The following elements
                are accepted:

                * `formula` : Chemical formula of the material.
                * `density` : Atomic density of the material [g / cm^3].

            mll_mat2 : A dictionary of the second MLL material. The following elements
                are accepted:

                * `formula` : Chemical formula of the material.
                * `density` : Atomic density of the material [g / cm^3].

            mll : A dictionary of multilayer Laue lens parameters. The following elements
                are accepted:

                * `n_min`, `n_max` : zone number of the first and the last layer.
                * `focus` : MLL's focal distance [um].
                * `mll_sigma` : Bilayer's interdiffusion length [um].
                * `mll_depth` : MLL's thickness [um].
                * `mll_wl` : Wavelength of the MLL [um].
        """
        super(MSParams, self).__init__(multislice=multislice, mll=mll,
                                       material1=mll_mat1.export_dict(),
                                       material2=mll_mat2.export_dict())
        self.mll_mat1, self.mll_mat2 = mll_mat1, mll_mat2

    @classmethod
    def _lookup_dict(cls) -> Dict[str, str]:
        lookup = {}
        for section in ('multislice', 'mll'):
            for option in cls.attr_dict[section]:
                lookup[option] = section
        return lookup

    @classmethod
    def import_default(cls, mll_mat1: Optional[Material]=None, mll_mat2: Optional[Material]=None,
                       **kwargs: Union[int, float]) -> MSParams:
        """Return the default :class:`MSParams`. Extra arguments override
        the default values if provided.

        Args:
            mll_mat1 : The first MLL material.
            mll_mat2 : The second MLL material.
            kwargs : Experimental parameters enlisted in :ref:`ms-parameters`.

        Returns:
            A :class:`MSParams` object with the default parameters.

        See Also:
            :ref:`ms-parameters` : Full list of experimental parameters.
        """
        return cls.import_ini(MS_PARAMETERS, mll_mat1, mll_mat2, **kwargs)

    @classmethod
    def import_ini(cls, ini_file: str, mll_mat1: Optional[Material]=None,
                   mll_mat2: Optional[Material]=None, **kwargs: Union[int, float]) -> MSParams:
        """Initialize a :class:`MSParams` object class with an ini file.

        Args:
            ini_file : Path to the ini file. Load the default parameters if
                None.
            mll_mat1 : The first MLL material. Initialized with `ini_file`
                if None.
            mll_mat2 : The second MLL material. Initialized with `ini_file`
                if None.
            kwargs : Experimental parameters enlisted in :ref:`ms-parameters`.
                Initialized with `ini_file` if not provided.

        Returns:
            A :class:`MSParams` object with all the attributes imported
            from the ini file.

        See Also:
            :ref:`ms-parameters` : Full list of experimental parameters.
        """
        attr_dict = cls._import_ini(ini_file)
        if mll_mat1 is None:
            mll_mat1 = Material(**attr_dict['material1'])
        if mll_mat2 is None:
            mll_mat2 = Material(**attr_dict['material2'])
        for option, section in cls._lookup_dict().items():
            if option in kwargs:
                attr_dict[section][option] = kwargs[option]
        return cls(multislice=attr_dict['multislice'], mll_mat1=mll_mat1,
                   mll_mat2=mll_mat2, mll=attr_dict['mll'])

    def get_mat1_r(self, energy: float) -> complex:
        """Return the Fresnel transmission coefficient of the first
        material in MLL bilayer.

        Args:
            energy : Beam's photon energy [keV].

        Returns:
            Fresnel transmission coefficient.
        """
        ref_idx = self.mll_mat1.get_ref_index(energy)
        return np.exp(2.0j * np.pi / self.wl * self.z_step * ref_idx)

    def get_mat2_r(self, energy: float) -> complex:
        """Return the Fresnel transmission coefficient of the second
        material in MLL bilayer.

        Args:
            Beam's photon energy [keV].

        Returns:
            Fresnel transmission coefficient.
        """
        ref_idx = self.mll_mat2.get_ref_index(energy)
        return np.exp(2.0j * np.pi / self.wl * self.z_step * ref_idx)

    def get_wavefront_size(self) -> int:
        """Return slice array size.

        Returns:
            Slice size.
        """
        return next_fast_len(int((self.x_max - self.x_min) // self.x_step))

    def get_xcoords(self, size: Optional[int]=None) -> np.ndarray:
        """Return a coordinate array of a slice.

        Args:
            size : Size of the array. Equals to :func:`MSParams.get_wavefront_size`
                if it's None.

        Returns:
            Coordinate array [um].
        """
        if size is None:
            size = self.get_wavefront_size()
        return (self.x_min + self.x_max) / 2 + self.x_step * np.arange(-size // 2, size // 2)

    def get_zcoords(self) -> np.ndarray:
        """Return a coordinate array along the propagation axis.

        Returns:
            Coordinate array [um].
        """
        return self.z_step * np.arange(self.mll_depth // self.z_step)

    def get_kernel(self, fx_arr: np.ndarray) -> np.ndarray:
        r"""Return diffraction kernel for the multislice propagation.

        Args:
            fx_arr : Spatial frequencies.

        Returns:
            Diffraction kernel.

        Notes:
            The diffraction kernel is given by:

            .. math::

                k(f_x, f_y) = \exp{-j \frac{2 \pi \Delta z}{\lambda}
                \sqrt{1 - \lambda^2 (f_x^2 + f_y^2)}}

            where :math:`\Delta z` --- slice thickness, :math:`f_x, f_y` ---
            spatial frequencies, and :math:`\lambda` --- wavelength.
        """
        kernel = np.exp(-2j * np.pi / self.wl * self.z_step * \
                        np.sqrt(1 - self.wl**2 * fx_arr.astype(complex)**2))
        return kernel
